from .libraries import sundials_cvode
from .sundials_matrix import *
from .sundials_linearsolver import *

# #include <sundials/sundials_direct.h>
# #include <sundials/sundials_iterative.h>
# #include <sundials/sundials_linearsolver.h>
# #include <sundials/sundials_matrix.h>
# #include <sundials/sundials_nvector.h>
#
# #ifdef __cplusplus  /* wrapper to enable C++ usage */
# extern "C" {
# #endif
#
#
# /*=================================================================
#   CVLS Constants
#   =================================================================*/
#
# #define CVLS_SUCCESS          0
# #define CVLS_MEM_NULL        -1
# #define CVLS_LMEM_NULL       -2
# #define CVLS_ILL_INPUT       -3
# #define CVLS_MEM_FAIL        -4
# #define CVLS_PMEM_NULL       -5
# #define CVLS_JACFUNC_UNRECVR -6
# #define CVLS_JACFUNC_RECVR   -7
# #define CVLS_SUNMAT_FAIL     -8
# #define CVLS_SUNLS_FAIL      -9
#
#
# /*=================================================================
#   CVLS user-supplied function prototypes
#   =================================================================*/
#
# typedef int (*CVLsJacFn)(realtype t, N_Vector y, N_Vector fy,
#                          SUNMatrix Jac, void *user_data,
#                          N_Vector tmp1, N_Vector tmp2, N_Vector tmp3);
#
# typedef int (*CVLsPrecSetupFn)(realtype t, N_Vector y, N_Vector fy,
#                                booleantype jok, booleantype *jcurPtr,
#                                realtype gamma, void *user_data);
#
# typedef int (*CVLsPrecSolveFn)(realtype t, N_Vector y, N_Vector fy,
#                                N_Vector r, N_Vector z, realtype gamma,
#                                realtype delta, int lr, void *user_data);
#
# typedef int (*CVLsJacTimesSetupFn)(realtype t, N_Vector y,
#                                    N_Vector fy, void *user_data);
#
# typedef int (*CVLsJacTimesVecFn)(N_Vector v, N_Vector Jv, realtype t,
#                                  N_Vector y, N_Vector fy,
#                                  void *user_data, N_Vector tmp);
#
#
# /*=================================================================
#   CVLS Exported functions
#   =================================================================*/
#
# SUNDIALS_EXPORT int CVodeSetLinearSolver(void *cvode_mem,
#                                          SUNLinearSolver LS,
#                                          SUNMatrix A);
CVodeSetLinearSolver = getattr(sundials_cvode, 'CVodeSetLinearSolver')
CVodeSetLinearSolver.argtypes = [c_void_p, SUNLinearSolver, SUNMatrix]
CVodeSetLinearSolver.restype = c_int
#
#
# /*-----------------------------------------------------------------
#   Optional inputs to the CVLS linear solver interface
#   -----------------------------------------------------------------*/
#
# SUNDIALS_EXPORT int CVodeSetJacFn(void *cvode_mem, CVLsJacFn jac);
# SUNDIALS_EXPORT int CVodeSetMaxStepsBetweenJac(void *cvode_mem,
#                                                long int msbj);
# SUNDIALS_EXPORT int CVodeSetEpsLin(void *cvode_mem, realtype eplifac);
# SUNDIALS_EXPORT int CVodeSetPreconditioner(void *cvode_mem,
#                                            CVLsPrecSetupFn pset,
#                                            CVLsPrecSolveFn psolve);
# SUNDIALS_EXPORT int CVodeSetJacTimes(void *cvode_mem,
#                                      CVLsJacTimesSetupFn jtsetup,
#                                      CVLsJacTimesVecFn jtimes);
#
# /*-----------------------------------------------------------------
#   Optional outputs from the CVLS linear solver interface
#   -----------------------------------------------------------------*/
#
# SUNDIALS_EXPORT int CVodeGetLinWorkSpace(void *cvode_mem,
#                                          long int *lenrwLS,
#                                          long int *leniwLS);
# SUNDIALS_EXPORT int CVodeGetNumJacEvals(void *cvode_mem,
#                                         long int *njevals);
# SUNDIALS_EXPORT int CVodeGetNumPrecEvals(void *cvode_mem,
#                                          long int *npevals);
# SUNDIALS_EXPORT int CVodeGetNumPrecSolves(void *cvode_mem,
#                                           long int *npsolves);
# SUNDIALS_EXPORT int CVodeGetNumLinIters(void *cvode_mem,
#                                         long int *nliters);
# SUNDIALS_EXPORT int CVodeGetNumLinConvFails(void *cvode_mem,
#                                             long int *nlcfails);
# SUNDIALS_EXPORT int CVodeGetNumJTSetupEvals(void *cvode_mem,
#                                               long int *njtsetups);
# SUNDIALS_EXPORT int CVodeGetNumJtimesEvals(void *cvode_mem,
#                                            long int *njvevals);
# SUNDIALS_EXPORT int CVodeGetNumLinRhsEvals(void *cvode_mem,
#                                            long int *nfevalsLS);
# SUNDIALS_EXPORT int CVodeGetLastLinFlag(void *cvode_mem,
#                                         long int *flag);
# SUNDIALS_EXPORT char *CVodeGetLinReturnFlagName(long int flag);



""" FMI 2.0 interface """

import pathlib
from ctypes import *
from . import free, calloc
from .fmi1 import _FMU, printLogMessage


fmi2Component            = c_void_p
fmi2ComponentEnvironment = c_void_p
fmi2FMUstate             = c_void_p
fmi2ValueReference       = c_uint
fmi2Real                 = c_double
fmi2Integer              = c_int
fmi2Boolean              = c_int
fmi2Char                 = c_char
fmi2String               = c_char_p
fmi2Type                 = c_int
fmi2Byte                 = c_char

fmi2Status = c_int

fmi2OK      = 0
fmi2Warning = 1
fmi2Discard = 2
fmi2Error   = 3
fmi2Fatal   = 4
fmi2Pending = 5

fmi2CallbackLoggerTYPE         = CFUNCTYPE(None, fmi2ComponentEnvironment, fmi2String, fmi2Status, fmi2String, fmi2String)
fmi2CallbackAllocateMemoryTYPE = CFUNCTYPE(c_void_p, c_size_t, c_size_t)
fmi2CallbackFreeMemoryTYPE     = CFUNCTYPE(None, c_void_p)
fmi2StepFinishedTYPE           = CFUNCTYPE(None, fmi2ComponentEnvironment, fmi2Status)

fmi2ModelExchange = 0
fmi2CoSimulation  = 1

fmi2True  = 1
fmi2False = 0

fmi2StatusKind = c_int

fmi2DoStepStatus       = 0
fmi2PendingStatus      = 1
fmi2LastSuccessfulTime = 2
fmi2Terminated         = 3


def allocateMemory(nobj, size):
    return calloc(nobj, size)


def freeMemory(obj):
    free(obj)


def stepFinished(componentEnvironment, status):
    pass


class fmi2CallbackFunctions(Structure):

    _fields_ = [('logger',               fmi2CallbackLoggerTYPE),
                ('allocateMemory',       fmi2CallbackAllocateMemoryTYPE),
                ('freeMemory',           fmi2CallbackFreeMemoryTYPE),
                ('stepFinished',         fmi2StepFinishedTYPE),
                ('componentEnvironment', fmi2ComponentEnvironment)]


class fmi2EventInfo(Structure):

    _fields_ = [('newDiscreteStatesNeeded',           fmi2Boolean),
                ('terminateSimulation',               fmi2Boolean),
                ('nominalsOfContinuousStatesChanged', fmi2Boolean),
                ('valuesOfContinuousStatesChanged',   fmi2Boolean),
                ('nextEventTimeDefined',              fmi2Boolean),
                ('nextEventTime',                     fmi2Real)]


class _FMU2(_FMU):
    """ Base class for FMI 2.0 FMUs """

    def __init__(self, **kwargs):

        super(_FMU2, self).__init__(**kwargs)

        # Inquire version numbers of header files and setting logging status
        self._fmi2Function('fmi2GetTypesPlatform', [], [], fmi2String)

        self._fmi2Function('fmi2GetVersion', [], [], fmi2String)

        self._fmi2Function('fmi2SetDebugLogging',
                           ['component', 'loggingOn', 'nCategories', 'categories'],
                           [fmi2Component, fmi2Boolean, c_size_t, POINTER(fmi2String)])

        # Creation and destruction of FMU instances and setting debug status
        self._fmi2Function('fmi2Instantiate',
                           ['instanceName', 'fmuType', 'guid', 'resourceLocation', 'callbacks', 'visible', 'loggingOn'],
                           [fmi2String, fmi2Type, fmi2String, fmi2String, POINTER(fmi2CallbackFunctions), fmi2Boolean, fmi2Boolean],
                           fmi2Component)

        self._fmi2Function('fmi2FreeInstance', ['component'], [fmi2Component], None)

        # Enter and exit initialization mode, terminate and reset
        self._fmi2Function('fmi2SetupExperiment',
                           ['component', 'toleranceDefined', 'tolerance', 'startTime', 'stopTimeDefined', 'stopTime'],
                           [fmi2Component, fmi2Boolean, fmi2Real, fmi2Real, fmi2Boolean, fmi2Real])

        self._fmi2Function('fmi2EnterInitializationMode', ['component'], [fmi2Component], fmi2Status)

        self._fmi2Function('fmi2ExitInitializationMode', ['component'], [fmi2Component], fmi2Status)

        self._fmi2Function('fmi2Terminate', ['component'], [fmi2Component], fmi2Status)

        self._fmi2Function('fmi2Reset', ['component'], [fmi2Component], fmi2Status)

        # Getting and setting variable values
        self._fmi2Function('fmi2GetReal',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Real)])

        self._fmi2Function('fmi2GetInteger',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Integer)])

        self._fmi2Function('fmi2GetBoolean',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Boolean)])

        self._fmi2Function('fmi2GetString',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2String)])

        self._fmi2Function('fmi2SetReal',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Real)])

        self._fmi2Function('fmi2SetInteger',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Integer)])

        self._fmi2Function('fmi2SetBoolean',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Boolean)])

        self._fmi2Function('fmi2SetString',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2String)])

        # Getting and setting the internal FMU state
        self._fmi2Function('fmi2GetFMUstate', ['component', 'FMUstate'],
                           [fmi2Component, POINTER(fmi2FMUstate)])

        self._fmi2Function('fmi2SetFMUstate', ['component', 'FMUstate'],
                           [fmi2Component, fmi2FMUstate])

        self._fmi2Function('fmi2FreeFMUstate', ['component', 'FMUstate'],
                           [fmi2Component, POINTER(fmi2FMUstate)])

        self._fmi2Function('fmi2SerializedFMUstateSize',
                           ['component', 'FMUstate', 'size'],
                           [fmi2Component, fmi2FMUstate, POINTER(c_size_t)])

        self._fmi2Function('fmi2SerializeFMUstate',
                           ['component', 'FMUstate', 'serializedState', 'size'],
                           [fmi2Component, fmi2FMUstate, POINTER(fmi2Byte), c_size_t])

        self._fmi2Function('fmi2DeSerializeFMUstate',
                           ['component', 'FMUstate', 'serializedState', 'size'],
                           [fmi2Component, POINTER(fmi2Byte), c_size_t, POINTER(fmi2FMUstate)])

        # Getting partial derivatives
        self._fmi2Function('fmi2GetDirectionalDerivative',
                           ['component', 'vUnknown_ref', 'nUnknown', 'vKnown_ref', 'nKnown', 'dvKnown', 'dvUnknown'],
                           [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Real), POINTER(fmi2Real)])

    def _fmi2Function(self, fname, argnames, argtypes, restype=fmi2Status):
        """ Add an FMI 2.0 function to this instance and add a wrapper that allows
        logging and checks the return code if the return type is fmi2Status

        Parameters:
            fname     the name of the function
            argnames  names of the arguments
            argtypes  types of the arguments
            restype   return type
        """

        if not hasattr(self.dll, fname):

            def raise_exception(*args):
                raise Exception("Function %s is missing in shared library." % fname)

            setattr(self, fname, raise_exception)

            return

        # get the exported function form the shared library
        f = getattr(self.dll, fname)
        f.argtypes = argtypes
        f.restype = restype

        def w(*args):
            """ Wrapper function for the FMI call """

            # call the FMI function
            res = f(*args)

            if self.fmiCallLogger is not None:
                # log the call
                self._log_fmi_args(fname, argnames, argtypes, args, restype, res)

            if restype == fmi2Status:  # status code
                # check the status code
                if res > fmi2Warning:
                    raise Exception("%s failed with status %d." % (fname, res))

            return res

        setattr(self, fname, w)

    # Inquire version numbers of header files and setting logging status

    def getTypesPlatform(self):
        types_platform = self.fmi2GetTypesPlatform()
        return types_platform.decode('utf-8')

    def getVersion(self):
        version = self.fmi2GetVersion()
        return version.decode('utf-8')

    def setDebugLogging(self, loggingOn, categories):
        categories_ = (fmi2String * len(categories))()
        categories_[:] = [c.encode('utf-8') for c in categories]
        self.fmi2SetDebugLogging(self.component, fmi2True if loggingOn else fmi2False, len(categories), categories_)

    # Creation and destruction of FMU instances and setting debug status

    def instantiate(self, visible=False, callbacks=None, loggingOn=False):

        kind = fmi2ModelExchange if isinstance(self, FMU2Model) else fmi2CoSimulation
        resourceLocation = pathlib.Path(self.unzipDirectory, 'resources').as_uri()

        if callbacks is None:
            callbacks = fmi2CallbackFunctions()
            callbacks.logger = fmi2CallbackLoggerTYPE(printLogMessage)
            callbacks.allocateMemory = fmi2CallbackAllocateMemoryTYPE(allocateMemory)
            callbacks.freeMemory = fmi2CallbackFreeMemoryTYPE(freeMemory)

        self.callbacks = callbacks

        self.component = self.fmi2Instantiate(self.instanceName.encode('utf-8'),
                                              kind,
                                              self.guid.encode('utf-8'),
                                              resourceLocation.encode('utf-8'),
                                              byref(self.callbacks),
                                              fmi2True if visible else fmi2False,
                                              fmi2True if loggingOn else fmi2False)

        if self.component is None:
            raise Exception("Failed to instantiate model")

    def freeInstance(self):
        self.fmi2FreeInstance(self.component)
        self.freeLibrary()

    # Enter and exit initialization mode, terminate and reset

    def setupExperiment(self, tolerance=None, startTime=0.0, stopTime=None):

        toleranceDefined = fmi2True if tolerance is not None else fmi2False

        if tolerance is None:
            tolerance = 0.0

        stopTimeDefined = fmi2True if stopTime is not None else fmi2False

        if stopTime is None:
            stopTime = 0.0

        return self.fmi2SetupExperiment(self.component, toleranceDefined, tolerance, startTime, stopTimeDefined, stopTime)

    def enterInitializationMode(self):
        return self.fmi2EnterInitializationMode(self.component)

    def exitInitializationMode(self):
        return self.fmi2ExitInitializationMode(self.component)

    def terminate(self):
        return self.fmi2Terminate(self.component)

    def reset(self):
        return self.fmi2Reset(self.component)

    # Getting and setting variable values

    def getReal(self, vr):
        vr = (fmi2ValueReference * len(vr))(*vr)
        value = (fmi2Real * len(vr))()
        self.fmi2GetReal(self.component, vr, len(vr), value)
        return list(value)

    def getInteger(self, vr):
        vr = (fmi2ValueReference * len(vr))(*vr)
        value = (fmi2Integer * len(vr))()
        self.fmi2GetInteger(self.component, vr, len(vr), value)
        return list(value)

    def getBoolean(self, vr):
        vr = (fmi2ValueReference * len(vr))(*vr)
        value = (fmi2Boolean * len(vr))()
        self.fmi2GetBoolean(self.component, vr, len(vr), value)
        return list(value)

    def getString(self, vr):
        vr = (fmi2ValueReference * len(vr))(*vr)
        value = (fmi2String * len(vr))()
        self.fmi2GetString(self.component, vr, len(vr), value)
        return list(value)

    def setReal(self, vr, value):
        vr = (fmi2ValueReference * len(vr))(*vr)
        value = (fmi2Real * len(vr))(*value)
        self.fmi2SetReal(self.component, vr, len(vr), value)

    def setInteger(self, vr, value):
        vr = (fmi2ValueReference * len(vr))(*vr)
        value = (fmi2Integer * len(vr))(*value)
        self.fmi2SetInteger(self.component, vr, len(vr), value)

    def setBoolean(self, vr, value):
        vr = (fmi2ValueReference * len(vr))(*vr)
        value = (fmi2Boolean * len(vr))(*value)
        self.fmi2SetBoolean(self.component, vr, len(vr), value)

    def setString(self, vr, value):
        vr = (fmi2ValueReference * len(vr))(*vr)
        value = map(lambda s: s.encode('utf-8') if s is not None else s, value)
        value = (fmi2String * len(vr))(*value)
        self.fmi2SetString(self.component, vr, len(vr), value)

    # Getting and setting the internal FMU state

    def getFMUstate(self):
        state = fmi2FMUstate()
        self.fmi2GetFMUstate(self.component, byref(state))
        return state

    def setFMUstate(self, state):
        self.fmi2SetFMUstate(self.component, state)

    def freeFMUstate(self, state):
        self.fmi2FreeFMUstate(self.component, byref(state))

    def serializeFMUstate(self, state):
        """ Serialize an FMU state

        Parameters:
            state   the FMU state

        Returns:
            the serialized state as a byte string
        """

        size = c_size_t()
        self.fmi2SerializedFMUstateSize(self.component, state, byref(size))
        serializedState = create_string_buffer(size.value)
        self.fmi2SerializeFMUstate(self.component, state, serializedState, size)
        return serializedState.raw

    def deSerializeFMUstate(self, serializedState, state=fmi2FMUstate()):
        """ De-serialize an FMU state

        Parameters:
            serializedState   the serialized state as a byte string
            state             previous FMU state for re-use (optional)
            
        Returns:
            the de-serialized FMU state
        """

        buffer = create_string_buffer(serializedState, size=len(serializedState))
        self.fmi2DeSerializeFMUstate(self.component, buffer, len(buffer), byref(state))
        return state

    # Getting partial derivatives

    def getDirectionalDerivative(self, vUnknown_ref, vKnown_ref, dvKnown):
        """ Get partial derivatives

        Parameters:
            vUnknown_ref    a list of value references of the unknowns
            vKnown_ref      a list of value references of the knowns
            dvKnown         a list of delta values (one per known)

        Returns:
            a list of the partial derivatives (one per unknown)
        """

        vUnknown_ref = (fmi2ValueReference * len(vUnknown_ref))(*vUnknown_ref)
        vKnown_ref = (fmi2ValueReference * len(vKnown_ref))(*vKnown_ref)
        dvKnown = (fmi2Real * len(dvKnown))(*dvKnown)
        dvUnknown = (fmi2Real * len(vUnknown_ref))()

        self.fmi2GetDirectionalDerivative(self.component, vUnknown_ref, len(vUnknown_ref), vKnown_ref, len(vKnown_ref), dvKnown, dvUnknown)

        return list(dvUnknown)


class FMU2Model(_FMU2):
    """ An FMI 2.0 Model Exchange FMU """

    def __init__(self, **kwargs):

        super(FMU2Model, self).__init__(**kwargs)

        # Enter and exit the different modes

        self._fmi2Function('fmi2EnterEventMode',
                           ['component'],
                           [fmi2Component])

        self._fmi2Function('fmi2NewDiscreteStates',
                           ['component', 'eventInfo'],
                           [fmi2Component, POINTER(fmi2EventInfo)])

        self._fmi2Function('fmi2EnterContinuousTimeMode',
                           ['component'],
                           [fmi2Component])

        self._fmi2Function('fmi2CompletedIntegratorStep',
                           ['component', 'noSetFMUStatePriorToCurrentPoint', 'enterEventMode', 'terminateSimulation'],
                           [fmi2Component, fmi2Boolean, POINTER(fmi2Boolean), POINTER(fmi2Boolean)])

        # Providing independent variables and re-initialization of caching

        self._fmi2Function('fmi2SetTime',
                           ['component', 'time'],
                           [fmi2Component, fmi2Real])

        self._fmi2Function('fmi2SetContinuousStates',
                           ['component', 'x', 'nx'],
                           [fmi2Component, POINTER(fmi2Real), c_size_t])

        # Evaluation of the model equations

        self._fmi2Function('fmi2GetDerivatives',
                           ['component', 'derivatives', 'nx'],
                           [fmi2Component, POINTER(fmi2Real), c_size_t])

        self._fmi2Function('fmi2GetEventIndicators',
                           ['component', 'eventIndicators', 'ni'],
                           [fmi2Component, POINTER(fmi2Real), c_size_t])

        self._fmi2Function('fmi2GetContinuousStates',
                           ['component', 'x', 'nx'],
                           [fmi2Component, POINTER(fmi2Real), c_size_t])

        self._fmi2Function('fmi2GetNominalsOfContinuousStates',
                           ['component', 'x_nominal', 'nx'],
                           [fmi2Component, POINTER(fmi2Real), c_size_t])

    # Enter and exit the different modes

    def enterEventMode(self):
        return self.fmi2EnterEventMode(self.component)

    def newDiscreteStates(self):

        eventInfo = fmi2EventInfo()

        self.fmi2NewDiscreteStates(self.component, byref(eventInfo))

        return (eventInfo.newDiscreteStatesNeeded           != fmi2False,
                eventInfo.terminateSimulation               != fmi2False,
                eventInfo.nominalsOfContinuousStatesChanged != fmi2False,
                eventInfo.valuesOfContinuousStatesChanged   != fmi2False,
                eventInfo.nextEventTimeDefined              != fmi2False,
                eventInfo.nextEventTime)

    def enterContinuousTimeMode(self):
        return self.fmi2EnterContinuousTimeMode(self.component)

    def completedIntegratorStep(self, noSetFMUStatePriorToCurrentPoint=fmi2True):
        enterEventMode = fmi2Boolean()
        terminateSimulation = fmi2Boolean()
        self.fmi2CompletedIntegratorStep(self.component, noSetFMUStatePriorToCurrentPoint, byref(enterEventMode), byref(terminateSimulation))
        return enterEventMode.value, terminateSimulation.value

    # Providing independent variables and re-initialization of caching

    def setTime(self, time):
        return self.fmi2SetTime(self.component, time)

    def setContinuousStates(self, x, nx):
        return self.fmi2SetContinuousStates(self.component, x, nx)

    # Evaluation of the model equations

    def getDerivatives(self, dx, nx):
        return self.fmi2GetDerivatives(self.component, dx, nx)

    def getEventIndicators(self, z, nz):
        return self.fmi2GetEventIndicators(self.component, z, nz)

    def getContinuousStates(self, x, nx):
        return self.fmi2GetContinuousStates(self.component, x, nx)

    def getNominalsOfContinuousStates(self, x_nominal, nx):
        return self.fmi2GetNominalsOfContinuousStates(self.component, x_nominal, nx)


class FMU2Slave(_FMU2):
    """ An FMI 2.0 Co-Simulation FMU """

    def __init__(self, instanceName=None, **kwargs):

        kwargs['instanceName'] = instanceName

        super(FMU2Slave, self).__init__(**kwargs)

        # Simulating the slave

        self._fmi2Function('fmi2SetRealInputDerivatives',
                           ['c', 'vr', 'nvr', 'order', 'value'],
                           [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Integer), POINTER(fmi2Real)])

        self._fmi2Function('fmi2GetRealOutputDerivatives',
                           ['c', 'vr', 'nvr', 'order', 'value'],
                           [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Integer), POINTER(fmi2Real)])

        self._fmi2Function('fmi2DoStep',
                           ['component', 'currentCommunicationPoint', 'communicationStepSize', 'noSetFMUStatePriorToCurrentPoint'],
                           [fmi2Component, fmi2Real, fmi2Real, fmi2Boolean])

        self._fmi2Function('fmi2CancelStep', ['component'], [fmi2Component])

        # Inquire slave status

        self._fmi2Function('fmi2GetStatus',
                           ['component', 'kind', 'value'],
                           [fmi2Component, fmi2StatusKind, POINTER(fmi2Status)])

        self._fmi2Function('fmi2GetRealStatus',
                           ['component', 'kind', 'value'],
                           [fmi2Component, fmi2StatusKind, POINTER(fmi2Real)])

        self._fmi2Function('fmi2GetIntegerStatus',
                           ['component', 'kind', 'value'],
                           [fmi2Component, fmi2StatusKind, POINTER(fmi2Integer)])

        self._fmi2Function('fmi2GetBooleanStatus',
                           ['component', 'kind', 'value'],
                           [fmi2Component, fmi2StatusKind, POINTER(fmi2Boolean)])

        self._fmi2Function('fmi2GetStringStatus',
                           ['component', 'kind', 'value'],
                           [fmi2Component, fmi2StatusKind, POINTER(fmi2String)])

    # Simulating the slave

    def setRealInputDerivatives(self, vr, order, value):
        vr = (fmi2ValueReference * len(vr))(*vr)
        order = (fmi2Integer * len(vr))(*order)
        value = (fmi2Real * len(vr))(*value)
        self.fmi2SetRealInputDerivatives(self.component, vr, len(vr), order, value)

    def getRealOutputDerivatives(self, vr, order):
        vr = (fmi2ValueReference * len(vr))(*vr)
        order = (fmi2Integer * len(vr))(*order)
        value = (fmi2Real * len(vr))()
        self.fmi2GetRealOutputDerivatives(self.component, vr, len(vr), order, value)
        return list(value)

    def doStep(self, currentCommunicationPoint, communicationStepSize, noSetFMUStatePriorToCurrentPoint=fmi2True):
        self.fmi2DoStep(self.component, currentCommunicationPoint, communicationStepSize, noSetFMUStatePriorToCurrentPoint)

    def cancelStep(self):
        self.fmi2CancelStep(self.component)

    # Inquire slave status

    def getStatus(self, kind):
        value = fmi2Status(fmi2OK)
        self.fmi2GetStatus(self.component, kind, byref(value))
        return value

    def getRealStatus(self, kind):
        value = fmi2Real(0.0)
        self.fmi2GetRealStatus(self.component, kind, byref(value))
        return value

    def getIntegerStatus(self, kind):
        value = fmi2Integer(0)
        self.fmi2GetIntegerStatus(self.component, kind, byref(value))
        return value

    def getBooleanStatus(self, kind):
        value = fmi2Boolean(fmi2False)
        self.fmi2GetBooleanStatus(self.component, kind, byref(value))
        return value

    def getStringStatus(self, kind):
        value = fmi2String(b'')
        self.fmi2GetStringStatus(self.component, kind, byref(value))
        return value



""" FMI 1.0 interface """

import os
import pathlib
from ctypes import *
from . import free, freeLibrary, platform, sharedLibraryExtension, calloc


fmi1Component      = c_void_p
fmi1ValueReference = c_uint
fmi1Real           = c_double
fmi1Integer        = c_int
fmi1Boolean        = c_char
fmi1String         = c_char_p

fmi1True  = b'\x01'
fmi1False = b'\x00'

fmi1UndefinedValueReference = -1

fmi1Status = c_int

fmi1OK      = 0
fmi1Warning = 1
fmi1Discard = 2
fmi1Error   = 3
fmi1Fatal   = 4

fmi1CallbackLoggerTYPE         = CFUNCTYPE(None, fmi1Component, fmi1String, fmi1Status, fmi1String, fmi1String)
fmi1CallbackAllocateMemoryTYPE = CFUNCTYPE(c_void_p, c_size_t, c_size_t)
fmi1CallbackFreeMemoryTYPE     = CFUNCTYPE(None, c_void_p)
# fmi1StepFinishedTYPE           = CFUNCTYPE(None, fmi1Component, fmi1Status)
fmi1StepFinishedTYPE           = c_void_p

fmi1StatusKind = c_int

fmi1DoStepStatus       = 0
fmi1PendingStatus      = 1
fmi1LastSuccessfulTime = 2


class fmi1CallbackFunctions(Structure):

    _fields_ = [('logger',         fmi1CallbackLoggerTYPE),
                ('allocateMemory', fmi1CallbackAllocateMemoryTYPE),
                ('freeMemory',     fmi1CallbackFreeMemoryTYPE),
                ('stepFinished',   fmi1StepFinishedTYPE)]

    def __str__(self):
        return 'fmi1CallbackFunctions(' \
               'logger=%s, ' \
               'allocateMemory=%s, ' \
               'freeMemory=%s, ' \
               'stepFinished=%s)' % (self.logger,
                                     self.allocateMemory,
                                     self.freeMemory,
                                     self.stepFinished)


class fmi1EventInfo(Structure):

    _fields_ = [('iterationConverged',          fmi1Boolean),
                ('stateValueReferencesChanged', fmi1Boolean),
                ('stateValuesChanged',          fmi1Boolean),
                ('terminateSimulation',         fmi1Boolean),
                ('upcomingTimeEvent',           fmi1Boolean),
                ('nextEventTime',               fmi1Real)]

    def __str__(self):
        return 'fmi1EventInfo(' \
               'iterationConverged=%s, ' \
               'stateValueReferencesChanged=%s, ' \
               'stateValuesChanged=%s, ' \
               'terminateSimulation=%s, ' \
               'upcomingTimeEvent=%s, ' \
               'nextEventTime=%s)' % (self.iterationConverged,
                                      self.stateValueReferencesChanged,
                                      self.stateValuesChanged,
                                      self.terminateSimulation,
                                      self.upcomingTimeEvent,
                                      self.nextEventTime)


def printLogMessage(component, instanceName, status, category, message):
    """ Print the FMU's log messages to the command line (works for both FMI 1.0 and 2.0) """

    label = ['OK', 'WARNING', 'DISCARD', 'ERROR', 'FATAL', 'PENDING'][status]
    print("[%s] %s" % (label, message.decode("utf-8")))


def allocateMemory(nobj, size):
    return calloc(nobj, size)


def freeMemory(obj):
    free(obj)


def stepFinished(componentEnvironment, status):
    pass


class _FMU(object):
    """ Base class for all FMUs """

    def __init__(self, guid, modelIdentifier, unzipDirectory, instanceName=None, libraryPath=None, fmiCallLogger=None):
        """
        Parameters:
            guid             the GUI from the modelDescription.xml
            modelIdentifier  the model identifier from the modelDescription.xml
            unzipDirectory   folder where the FMU has been extracted
            instanceName     the name of the FMU instance
            libraryPath      path to the shared library
            fmiCallLogger    logger callback that takes a message as input
        """

        self.guid = guid
        self.modelIdentifier = modelIdentifier
        self.unzipDirectory = unzipDirectory
        self.instanceName = instanceName if instanceName is not None else self.modelIdentifier
        self.fmiCallLogger = fmiCallLogger

        # remember the current working directory
        work_dir = os.getcwd()

        if libraryPath is None:
            library_dir = os.path.join(unzipDirectory, 'binaries', platform)
            libraryPath = str(os.path.join(library_dir, self.modelIdentifier + sharedLibraryExtension))
        else:
            library_dir = os.path.dirname(libraryPath)

        # check if shared library exists
        if not os.path.isfile(libraryPath):
            raise Exception("Cannot find shared library %s." % libraryPath)

        # change to the library directory as some DLLs expect this to resolve dependencies
        os.chdir(library_dir)

        # load the shared library
        try:
            self.dll = cdll.LoadLibrary(libraryPath)
        except Exception as e:
            raise Exception("Failed to load shared library %s. %s" % (libraryPath, e))

        # change back to the working directory
        os.chdir(work_dir)

        self.component = None

        self.callbacks = None
        " Reference to the callbacks struct (to save it from GC)"

    def freeLibrary(self):
        # unload the shared library
        freeLibrary(self.dll._handle)

    def _log_fmi_args(self, fname, argnames, argtypes, args, restype, res):
        """ Format FMI arguments and pass them to the logger """

        f = fname + '('

        l = []

        for i, (n, t, v) in enumerate(zip(argnames, argtypes, args)):

            a = n + '='

            if t == c_void_p:
                if isinstance(v, c_void_p):
                    v = v.value
                a += hex(0 if v is None else v)
            elif t == c_bool:
                a += str(v.value)
            elif t == POINTER(c_uint):
                # value references
                if v is None:
                    a += 'NULL'
                else:
                    a += '[' + ', '.join(map(str, v)) + ']'
            elif t == POINTER(c_double):
                if hasattr(v, '__len__'):
                    # c_double_Array_N
                    a += '[' + ', '.join(map(str, v)) + ']'
                else:
                    if len(args) > i + 1:
                        # double pointers are always flowed by the size of the array
                        arr = v[:args[i + 1]]
                        a += '[' + ', '.join(map(str, arr)) + ']'
                    else:
                        # except for fmi3DoStep
                        v_ = cast(v, POINTER(c_double))
                        a += str(str(v_.contents.value))

            elif hasattr(v, '_obj'):
                # byref object
                if hasattr(v._obj, 'value'):
                    # pointer (e.g. c_char_p)
                    a += str(v._obj.value)
                else:
                    # struct
                    a += str(v._obj)
            elif hasattr(v, 'decode'):
                # UTF-8 byte string
                a += '"' + v.decode('utf-8') + '"'
            else:
                a += str(v)

            l.append(a)

        f += ', '.join(l) + ')'

        if restype == c_int:

            f += ' -> '

            if res == 0:
                f += 'OK'
            elif res == 1:
                f += 'WARNING'
            elif res == 2:
                f += 'DISCARD'
            elif res == 3:
                f += 'ERROR'
            elif res == 4:
                f += 'FATAL'
            elif res == 5:
                f += 'PENDING'
            else:
                f += str(res)
        elif restype == c_void_p:
            f += ' -> ' + hex(0 if res is None else res)

        self.fmiCallLogger(f)


class _FMU1(_FMU):
    """ Base class for FMI 1.0 FMUs """

    def __init__(self, **kwargs):

        super(_FMU1, self).__init__(**kwargs)

        # Inquire version numbers of header files
        self._fmi1Function('GetVersion', [], [], fmi1String)

        self._fmi1Function('SetDebugLogging', ['component', 'loggingOn'], [fmi1Component, fmi1Boolean], fmi1Status)

        # Data Exchange Functions
        self._fmi1Function('GetReal',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Real)])

        self._fmi1Function('GetInteger',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Integer)])

        self._fmi1Function('GetBoolean',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Boolean)])

        self._fmi1Function('GetString',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1String)])

        self._fmi1Function('SetReal',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Real)])

        self._fmi1Function('SetInteger',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Integer)])

        self._fmi1Function('SetBoolean',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Boolean)])

        self._fmi1Function('SetString',
                           ['component', 'vr', 'nvr', 'value'],
                           [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1String)])

    def _fmi1Function(self, fname, argnames, argtypes, restype=fmi1Status):
        """ Add an FMI 1.0 function to this instance and add a wrapper that allows
        logging and checks the return code if the return type if fmi1Status

        Parameters:
            fname     the name of the function (without 'fmi' prefix)
            argnames  names of the arguments
            argtypes  types of the arguments
            restype   return type
        """

        # get the exported function form the shared library
        f = getattr(self.dll, self.modelIdentifier + '_fmi' + fname)
        f.argtypes = argtypes
        f.restype = restype

        def w(*args):
            """ Wrapper function for the FMI call """

            # call the FMI function
            res = f(*args)

            if self.fmiCallLogger is not None:
                # log the call
                self._log_fmi_args('fmi' + fname, argnames, argtypes, args, restype, res)

            if restype == fmi1Status:
                # check the status code
                if res > fmi1Warning:
                    raise Exception("fmi%s failed with status %d." % (fname, res))

            return res

        # add the function to the instance
        setattr(self, 'fmi1' + fname, w)

    # Inquire version numbers of header files

    def getVersion(self):
        version = self.fmi1GetVersion()
        return version.decode('utf-8')

    def setDebugLogging(self, loggingOn):
        self.fmi1SetDebugLogging(self.component, fmi1True if loggingOn else fmi1False)

    # Data Exchange Functions

    def getReal(self, vr):
        vr = (fmi1ValueReference * len(vr))(*vr)
        value = (fmi1Real * len(vr))()
        self.fmi1GetReal(self.component, vr, len(vr), value)
        return list(value)

    def getInteger(self, vr):
        vr = (fmi1ValueReference * len(vr))(*vr)
        value = (fmi1Integer * len(vr))()
        self.fmi1GetInteger(self.component, vr, len(vr), value)
        return list(value)

    def getBoolean(self, vr):
        vr = (fmi1ValueReference * len(vr))(*vr)
        value = (fmi1Boolean * len(vr))()
        self.fmi1GetBoolean(self.component, vr, len(vr), value)
        return list(map(lambda b: 0 if b == fmi1False else 1, value))

    def getString(self, vr):
        vr = (fmi1ValueReference * len(vr))(*vr)
        value = (fmi1String * len(vr))()
        self.fmi1GetString(self.component, vr, len(vr), value)
        return list(value)

    def setReal(self, vr, value):
        vr = (fmi1ValueReference * len(vr))(*vr)
        value = (fmi1Real * len(vr))(*value)
        self.fmi1SetReal(self.component, vr, len(vr), value)

    def setInteger(self, vr, value):
        vr = (fmi1ValueReference * len(vr))(*vr)
        value = (fmi1Integer * len(vr))(*value)
        self.fmi1SetInteger(self.component, vr, len(vr), value)

    def setBoolean(self, vr, value):
        # convert value to a byte string
        s = b''
        for v in value:
            s += fmi1True if v else fmi1False

        vr = (fmi1ValueReference * len(vr))(*vr)
        value = (fmi1Boolean * len(vr))(s)
        self.fmi1SetBoolean(self.component, vr, len(vr), value)

    def setString(self, vr, value):
        vr = (fmi1ValueReference * len(vr))(*vr)
        value = map(lambda s: s.encode('utf-8'), value)
        value = (fmi1String * len(vr))(*value)
        self.fmi1SetString(self.component, vr, len(vr), value)


class FMU1Slave(_FMU1):
    """ An FMI 1.0 Co-Simulation FMU """

    def __init__(self, **kwargs):

        super(FMU1Slave, self).__init__(**kwargs)

        # Inquire version numbers of header files
        self._fmi1Function('GetTypesPlatform', [], [], fmi1String)

        # Creation and destruction of slave instances and setting debug status
        self._fmi1Function('InstantiateSlave',
                           ['instanceName', 'guid', 'fmuLocation', 'mimeType', 'timeout', 'visible', 'interactive', 'functions', 'loggingOn'],
                           [fmi1String, fmi1String, fmi1String, fmi1String, fmi1Real, fmi1Boolean, fmi1Boolean, fmi1CallbackFunctions, fmi1Boolean],
                           fmi1Component)

        self._fmi1Function('InitializeSlave',
                           ['component', 'tStart', 'stopTimeDefined', 'tStop'],
                           [fmi1Component, fmi1Real, fmi1Boolean, fmi1Real])

        self._fmi1Function('TerminateSlave', ['component'], [fmi1Component], fmi1Status)

        self._fmi1Function('ResetSlave', ['component'], [fmi1Component], fmi1Status)

        self._fmi1Function('FreeSlaveInstance', ['component'], [fmi1Component], None)

        self._fmi1Function('SetRealInputDerivatives',
                           ['c', 'vr', 'nvr', 'order', 'value'],
                           [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Integer), POINTER(fmi1Real)])

        self._fmi1Function('GetRealOutputDerivatives',
                           ['c', 'vr', 'nvr', 'order', 'value'],
                           [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Integer), POINTER(fmi1Real)])

        self._fmi1Function('CancelStep', ['component'], [fmi1Component])

        self._fmi1Function('DoStep',
                           ['component', 'currentCommunicationPoint', 'communicationStepSize', 'newStep'],
                           [fmi1Component, fmi1Real, fmi1Real, fmi1Boolean])

        self._fmi1Function('GetStatus',
                           ['component', 'kind', 'value'],
                           [fmi1Component, fmi1StatusKind, POINTER(fmi1Status)])

        self._fmi1Function('GetRealStatus',
                           ['component', 'kind', 'value'],
                           [fmi1Component, fmi1StatusKind, POINTER(fmi1Real)])

        self._fmi1Function('GetIntegerStatus',
                           ['component', 'kind', 'value'],
                           [fmi1Component, fmi1StatusKind, POINTER(fmi1Integer)])

        self._fmi1Function('GetBooleanStatus',
                           ['component', 'kind', 'value'],
                           [fmi1Component, fmi1StatusKind, POINTER(fmi1Boolean)])

        self._fmi1Function('GetStringStatus',
                           ['component', 'kind', 'value'],
                           [fmi1Component, fmi1StatusKind, POINTER(fmi1String)])

    # Creation and destruction of slave instances and setting debug status

    def instantiate(self, mimeType='application/x-fmu-sharedlibrary', timeout=0, visible=fmi1False,
                    interactive=fmi1False, functions=None, loggingOn=False):

        fmuLocation = pathlib.Path(self.unzipDirectory).as_uri()

        if functions is None:
            functions = fmi1CallbackFunctions()
            functions.logger = fmi1CallbackLoggerTYPE(printLogMessage)
            functions.allocateMemory = fmi1CallbackAllocateMemoryTYPE(allocateMemory)
            functions.freeMemory = fmi1CallbackFreeMemoryTYPE(freeMemory)
            functions.stepFinished = None

        self.callbacks = functions

        self.component = self.fmi1InstantiateSlave(self.instanceName.encode('UTF-8'),
                                                   self.guid.encode('UTF-8'),
                                                   fmuLocation.encode('UTF-8'),
                                                   mimeType.encode('UTF-8'),
                                                   timeout,
                                                   visible,
                                                   interactive,
                                                   functions,
                                                   fmi1True if loggingOn else fmi1False)

    # Inquire version numbers of header files

    def getTypesPlatform(self):
        types_platform = self.fmi1GetTypesPlatform()
        return types_platform.decode('utf-8')

    # Creation and destruction of slave instances and setting debug status

    def initialize(self, tStart=0.0, stopTime=None):
        stopTimeDefined = fmi1True if stopTime is not None else fmi1False
        tStop = stopTime if stopTime is not None else 0.0
        return self.fmi1InitializeSlave(self.component, tStart, stopTimeDefined, tStop)

    def terminate(self):
        return self.fmi1TerminateSlave(self.component)

    def reset(self):
        return self.fmi1ResetSlave(self.component)

    def freeInstance(self):
        self.fmi1FreeSlaveInstance(self.component)
        self.freeLibrary()

    def setRealInputDerivatives(self, vr, order, value):
        vr = (fmi1ValueReference * len(vr))(*vr)
        order = (fmi1Integer * len(vr))(*order)
        value = (fmi1Real * len(vr))(*value)
        self.fmi1SetRealInputDerivatives(self.component, vr, len(vr), order, value)

    def getRealOutputDerivatives(self, vr, order):
        vr = (fmi1ValueReference * len(vr))(*vr)
        order = (fmi1Integer * len(vr))(*order)
        value = (fmi1Real * len(vr))()
        self.fmi1GetRealOutputDerivatives(self.component, vr, len(vr), order, value)
        return list(value)

    def cancelStep(self):
        return self.fmi1CancelStep(self.component)

    def doStep(self, currentCommunicationPoint, communicationStepSize, newStep=fmi1True):
        self.fmi1DoStep(self.component, currentCommunicationPoint, communicationStepSize, newStep)

    def getStatus(self, kind):
        value = fmi1Status(fmi1OK)
        self.fmi1GetStatus(self.component, kind, byref(value))
        return value

    def getRealStatus(self, kind):
        value = fmi1Real(0.0)
        self.fmi1GetRealStatus(self.component, kind, byref(value))
        return value

    def getIntegerStatus(self, kind):
        value = fmi1Integer(0)
        self.fmi1GetIntegerStatus(self.component, kind, byref(value))
        return value

    def getBooleanStatus(self, kind):
        value = fmi1Boolean(fmi1False)
        self.fmi1GetBooleanStatus(self.component, kind, byref(value))
        return value

    def getStringStatus(self, kind):
        value = fmi1String(b'')
        self.fmi1GetStringStatus(self.component, kind, byref(value))
        return value


class FMU1Model(_FMU1):
    """ An FMI 1.0 Model Exchange FMU """

    def __init__(self, **kwargs):

        super(FMU1Model, self).__init__(**kwargs)

        # Inquire version numbers of header files
        self._fmi1Function('GetModelTypesPlatform', [], [], fmi1String)

        # Creation and destruction of model instances and setting debug status
        self._fmi1Function('InstantiateModel',
                           ['instanceName', 'guid', 'functions', 'loggingOn'],
                           [fmi1String, fmi1String, fmi1CallbackFunctions, fmi1Boolean],
                           fmi1Component)

        self._fmi1Function('FreeModelInstance', ['component'], [fmi1Component], None)

        # Providing independent variables and re-initialization of caching
        self._fmi1Function('SetTime', ['component', 'time'], [fmi1Component, fmi1Real])

        self._fmi1Function('SetContinuousStates',
                           ['component', 'x', 'nx'],
                           [fmi1Component, POINTER(fmi1Real), c_size_t])

        self._fmi1Function('CompletedIntegratorStep',
                           ['component', 'callEventUpdate'],
                           [fmi1Component, POINTER(fmi1Boolean)])

        # Evaluation of the model equations
        self._fmi1Function('Initialize',
                           ['component', 'toleranceControlled', 'relativeTolerance', 'eventInfo'],
                           [fmi1Component, fmi1Boolean, fmi1Real, POINTER(fmi1EventInfo)])

        self._fmi1Function('GetDerivatives',
                           ['component', 'derivatives', 'nx'],
                           [fmi1Component, POINTER(fmi1Real), c_size_t])

        self._fmi1Function('GetEventIndicators',
                           ['component', 'eventIndicators', 'ni'],
                           [fmi1Component, POINTER(fmi1Real), c_size_t])

        self._fmi1Function('EventUpdate',
                           ['component', 'intermediateResults', 'eventInfo'],
                           [fmi1Component, fmi1Boolean, POINTER(fmi1EventInfo)])

        self._fmi1Function('GetContinuousStates',
                           ['component', 'states', 'nx'],
                           [fmi1Component, POINTER(fmi1Real), c_size_t])

        self._fmi1Function('GetNominalContinuousStates',
                           ['component', 'x_nominal', 'nx'],
                           [fmi1Component, POINTER(fmi1Real), c_size_t])

        self._fmi1Function('GetStateValueReferences',
                           ['component', 'x_nominal', 'nx'],
                           [fmi1Component, POINTER(fmi1ValueReference), c_size_t])

        self._fmi1Function('Terminate', ['component'], [fmi1Component])

    # Inquire version numbers of header files

    def getTypesPlatform(self):
        types_platform = self.fmi1GetModelTypesPlatform()
        return types_platform.decode('utf-8')

    # Creation and destruction of model instances and setting debug status

    def instantiate(self, functions=None, loggingOn=False):

        if functions is None:
            functions = fmi1CallbackFunctions()
            functions.logger = fmi1CallbackLoggerTYPE(printLogMessage)
            functions.allocateMemory = fmi1CallbackAllocateMemoryTYPE(allocateMemory)
            functions.freeMemory = fmi1CallbackFreeMemoryTYPE(freeMemory)
            functions.stepFinished = None

        self.callbacks = functions

        self.component = self.fmi1InstantiateModel(self.instanceName.encode('UTF-8'),
                                                   self.guid.encode('UTF-8'),
                                                   self.callbacks,
                                                   fmi1True if loggingOn else fmi1False)

        if self.component is None:
            raise Exception("Failed to instantiate model")

    def freeInstance(self):
        self.fmi1FreeModelInstance(self.component)
        self.freeLibrary()

    # Providing independent variables and re-initialization of caching

    def setTime(self, time):
        return self.fmi1SetTime(self.component, time)

    def setContinuousStates(self, states, size):
        return self.fmi1SetContinuousStates(self.component, states, size)

    def completedIntegratorStep(self):
        stepEvent = fmi1Boolean()
        self.fmi1CompletedIntegratorStep(self.component, byref(stepEvent))
        return stepEvent.value != fmi1False

    # Evaluation of the model equations

    def initialize(self, toleranceControlled=fmi1False, relativeTolerance=0.0):
        eventInfo = fmi1EventInfo()
        self.fmi1Initialize(self.component, toleranceControlled, relativeTolerance, byref(eventInfo))
        return (eventInfo.iterationConverged != fmi1False,
                eventInfo.stateValueReferencesChanged != fmi1False,
                eventInfo.stateValuesChanged != fmi1False,
                eventInfo.terminateSimulation != fmi1False,
                eventInfo.upcomingTimeEvent != fmi1False,
                eventInfo.nextEventTime)

    def getDerivatives(self, derivatives, size):
        return self.fmi1GetDerivatives(self.component, derivatives, size)

    def getEventIndicators(self, eventIndicators, size):
        return self.fmi1GetEventIndicators(self.component, eventIndicators, size)

    def eventUpdate(self, intermediateResults=fmi1False):
        eventInfo = fmi1EventInfo()
        self.fmi1EventUpdate(self.component, intermediateResults, byref(eventInfo))
        return (eventInfo.iterationConverged          != fmi1False,
                eventInfo.stateValueReferencesChanged != fmi1False,
                eventInfo.stateValuesChanged          != fmi1False,
                eventInfo.terminateSimulation         != fmi1False,
                eventInfo.upcomingTimeEvent           != fmi1False,
                eventInfo.nextEventTime)

    def getContinuousStates(self, states, size):
        return self.fmi1GetContinuousStates(self.component, states, size)

    def getNominalContinuousStates(self, x_nominal, size):
        return self.fmi1GetNominalContinuousStates(self.component, x_nominal, size)

    def getStateValueReferences(self, vrx, size):
        return self.fmi1GetStateValueReferences(self.component, vrx, size)

    def terminate(self):
        return self.fmi1Terminate(self.component)



""" Validation of the modelDescription.xml """

from typing import List

from fmpy.model_description import ModelDescription


def validate_fmu(filename: str) -> List[str]:
    """ Validate the following aspects of an FMU

    modelDescription.xml:
    - validation against the XML schema
    - uniqueness and validity of variable names
    - completeness and integrity of the ModelStructure
    - required start values
    - combinations of causality and variability
    - units

    Parameters:
        filename  filename of the FMU, directory with extracted FMU or file like object

    Returns:
        a list of the problems found
    """

    from . import read_model_description
    from .model_description import ValidationError
    import os
    import zipfile

    problems = []

    # check file paths
    if isinstance(filename, str) and os.path.isfile(filename):
        with zipfile.ZipFile(filename, 'r') as zf:
            for file in zf.filelist:
                if '\\' in file.orig_filename:
                    problems.append(f'The file path "{file.orig_filename}" contains a backslash.')

    try:
        read_model_description(filename,
                               validate=True,
                               validate_variable_names=True,
                               validate_model_structure=True)
    except ValidationError as e:
        problems = e.problems
    except Exception as e:
        problems = [str(e)]

    return problems


def validate_model_description(model_description: ModelDescription, validate_variable_names: bool = False, validate_model_structure: bool = False) -> List[str]:

    problems = []

    if validate_variable_names:
        problems += _validate_variable_names(model_description)

    unit_definitions = {}

    for unit in model_description.unitDefinitions:
        unit_definitions[unit.name] = [display_unit.name for display_unit in unit.displayUnits]

    variable_names = set()

    # assert unique variable names (FMI 1.0 spec, p. 34, FMI 2.0 spec, p. 45)
    for v in model_description.modelVariables:
        if v.name in variable_names:
            problems.append(f'The variable name "{v.name}" (line {v.sourceline}) is not unique.')
        variable_names.add(v.name)

    is_fmi2 = model_description.fmiVersion == '2.0'
    is_fmi3 = model_description.fmiVersion.startswith('3.0')

    if is_fmi2 or is_fmi3:

        # assert required start values (see FMI 2.0 spec, p. 53)
        for v in model_description.modelVariables:
            if v.type != 'Clock' and (v.initial in {'exact', 'approx'} or v.causality == 'input') and v.start is None:
                problems.append(f'Variable "{v.name}" (line {v.sourceline}) has no start value.')

        # assert that initial is not set for input and independent variables (see FMI 2.0 spec, p. 49)
        for v in model_description.modelVariables:
            if v.causality in {'input', 'independent'} and v.initial is not None:
                problems.append(f'Variable "{v.name}" (line {v.sourceline}) " has causality "{v.causality}" but defines a intial "{v.initial}".')

        # legal combinations of causality and variability (see FMI 2.0 spec, p. 49)
        legal_combinations = {
            ('parameter', 'fixed'),
            ('parameter', 'tunable'),
            ('calculatedParameter', 'fixed'),
            ('calculatedParameter', 'tunable'),
            ('structuralParameter', 'fixed'),
            ('structuralParameter', 'tunable'),
            ('input', 'discrete'),
            ('input', 'continuous'),
            ('output', 'constant'),
            ('output', 'discrete'),
            ('output', 'continuous'),
            ('local', 'constant'),
            ('local', 'fixed'),
            ('local', 'tunable'),
            ('local', 'discrete'),
            ('local', 'continuous'),
            ('independent', 'continuous'),
        }

        if is_fmi3:
            legal_combinations.add(('input', 'clock'))
            legal_combinations.add(('output', 'clock'))

        for v in model_description.modelVariables:
            if (v.causality, v.variability) not in legal_combinations:
                problems.append(f'The combination causality="{v.causality}" and variability="{v.variability}" '
                                f'in variable "{v.name}" (line {v.sourceline}) is not allowed.')

        # check for illegal start values (see FMI 2.0.2 spec, p. 49)
        for v in model_description.modelVariables:

            if v.initial == 'calculated' and v.start:
                problems.append(f'The variable "{v.name}" (line {v.sourceline}) has initial="calculated" but provides a start value.')

            if v.causality in 'independent' and v.start:
                problems.append(f'The variable "{v.name}" (line {v.sourceline}) has causality="independent" but provides a start value.')

        # validate units (see FMI 2.0 spec, p. 33ff.)
        for v in model_description.modelVariables:

            unit = v.unit

            if unit is None and v.declaredType is not None:
                unit = v.declaredType.unit

            if unit is not None and unit not in unit_definitions:
                problems.append(f'The unit "{unit}" of variable "{v.name}" (line {v.sourceline}) is not defined.')

            if v.displayUnit is not None and v.displayUnit not in unit_definitions[unit]:
                problems.append(f'The display unit "{v.displayUnit}" of variable "{v.name}" (line {v.sourceline}) is not defined.')

        if validate_model_structure:
            problems += _validate_model_structure(model_description)

    if is_fmi3:

        # assert independent variable
        if sum(v.causality == 'independent' for v in model_description.modelVariables) != 1:
            problems.append("Exactly one independent variable must be defined.")

        # assert unique value references
        variables = dict()
        
        for v in model_description.modelVariables:
            if v.valueReference in variables:
                p = variables[v.valueReference]
                problems.append(f'Variable "{v.name}" (line {v.sourceline}) has the same value reference as variable "{p.name}" (line {p.sourceline}).')
            else:
                variables[v.valueReference] = v

    return problems


def _validate_model_structure(model_description: ModelDescription) -> List[str]:

    problems = []

    # validate outputs
    expected_outputs = set(v for v in model_description.modelVariables if v.causality == 'output')
    outputs = set(u.variable for u in model_description.outputs)

    if expected_outputs != outputs:
        problems.append('ModelStructure/Outputs must have exactly one entry for each variable with causality="output".')

    # validate derivatives
    derivatives = set(v for v in model_description.modelVariables if v.derivative is not None)
    for i, state_derivative in enumerate(model_description.derivatives):
        if state_derivative.variable not in derivatives:
            problems.append(f'The variable "{state_derivative.variable.name}" (line {state_derivative.variable.sourceline}) '
                            f'referenced by the continuous state derivative {i + 1} (line {state_derivative.sourceline}) '
                            f'must have the attribute "derivative".')

    # validate initial unknowns
    expected_initial_unknowns = set()

    for variable in model_description.modelVariables:

        if variable.causality == 'output' and variable.initial in {'approx', 'calculated'} and not variable.clocks:
            expected_initial_unknowns.add(variable)

        if variable.causality == 'calculatedParameter':
            expected_initial_unknowns.add(variable)

    try:
        for unknown in model_description.derivatives:
            derivative = unknown.variable
            state = derivative.derivative
            for variable in [state, derivative]:
                if variable.initial in {'approx', 'calculated'}:
                    expected_initial_unknowns.add(variable)

        initial_unknowns = set(v.variable for v in model_description.initialUnknowns)

        if initial_unknowns != expected_initial_unknowns:
            expected = ', '.join(sorted(v.name for v in expected_initial_unknowns))
            actual = ', '.join(sorted(v.name for v in initial_unknowns))
            problem = ("ModelStructure/InitialUnknowns does not contain the expected set of variables." 
                       f" Expected {{{ expected }}} but was {{{ actual }}}.")
            problems.append(problem)
    except:
        pass  # this check may fail due to inconsistencies detected above

    return problems


def _validate_variable_names(model_description: ModelDescription) -> List[str]:

    problems = []

    if model_description.variableNamingConvention == 'flat':

        for variable in model_description.modelVariables:

            if u'\u000D' in variable.name:
                problems.append(f'Variable "{variable.name}" (line {variable.sourceline}) contains an illegal carriage return character (U+000D).')

            if u'\u000A' in variable.name:
                problems.append(f'Variable "{variable.name}" (line {variable.sourceline}) contains an illegal line feed character (U+000A).')

            if u'\u0009' in variable.name:
                problems.append(f'Variable "{variable.name}" (line {variable.sourceline}) contains an illegal tab character (U+0009).')

    else:  # variableNamingConvention == structured

        from lark import Lark

        grammar = r"""
            name            : identifier | "der(" identifier ("," unsignedinteger)? ")"
            identifier      : bname arrayindices? ("." bname arrayindices?)*
            bname           : nondigit (nondigit|digit)* | qname
            nondigit        : "_" | "a".."z" | "A".."Z"
            digit           : "0".."9"
            qname           : "'" ( qchar | escape ) ( qchar | escape ) "'"
            qchar           : nondigit | digit | "!" | "#" | "$" | "%" | "&" | "(" | ")" 
                              | "*" | "+" | "," | "-" | "." | "/" | ":" | ";" | "<" | ">"
                              | "=" | "?" | "@" | "[" | "]" | "^" | "{" | "}" | "|" | "~" | " "
            escape          : "\'" | "\"" | "\?" | "\\" | "\a" | "\b" | "\f" | "\n" | "\r" | "\t" | "\v"
            arrayindices    : "[" unsignedinteger ("," unsignedinteger)* "]"
            unsignedinteger : digit+
            """

        parser = Lark(grammar, start='name')

        for variable in model_description.modelVariables:
            try:
                parser.parse(variable.name)
            except Exception as e:
                problems.append('"%s" (line %s) is not a legal variable name for naming convention "structured". %s'
                                % (variable.name, variable.sourceline, e))

    return problems




def get_vendor_ids(tools_csv):
    """ Get vendor and tool info from tools.csv

    Parameters:
        tools_csv  path to the _data/tools.csv in the fmi-standard.org repository

    Returns:
        a dictionary {vendor_id: (tool_id, tool_name)}
    """

    import csv

    vendors = {}

    with open(tools_csv, 'r') as csvfile:

        reader = csv.reader(csvfile, delimiter=',', quotechar='"')

        next(reader)  # skip the header

        for row in reader:

            tool_name, tool_id, vendor_id = row[:3]

            if vendor_id in vendors:
                vendors[vendor_id].append((tool_id, tool_name))
            else:
                vendors[vendor_id] = [(tool_id, tool_name)]

    return vendors


def validate_result(result, reference, t_start, t_stop):
    """ Validate a simulation result against a reference result

    Parameters:
        result      structured NumPy array where the first column is the time
        reference   same as result
        ...

    Returns:
        problems    a list of problems
    """

    import numpy as np

    t_ref = reference[reference.dtype.names[0]]
    t_res = result[result.dtype.names[0]]

    # at least two samples are required
    if result.size < 2:
        return 'The result must have at least two samples'

    # check if stop time has been reached
    if t_res[0] > t_start:
        return 'The result starts at %g after the start time (%g s)' % (t_res[0], t_start)

    # check if stop time has been reached
    if t_res[-1] < t_stop:
        return 'The result ends at %g s before the stop time (%g s)' % (t_res[-1], t_stop)

    # check if all reference signals are contained in the result
    for name in reference.dtype.names[1:]:
        if name not in result.dtype.names:
            return 'Variable "%s" is missing' % name

    # find the signal with the most outliers
    for name in result.dtype.names[1:]:

        if name not in reference.dtype.names:
            continue

        y_ref = reference[name]
        y_res = result[name]
        _, _, _, outliers = validate_signal(t=t_res, y=y_res, t_ref=t_ref, y_ref=y_ref, t_start=t_start, t_stop=t_stop)

        # calculate the relative number of outliers
        rel_out = np.sum(outliers) / float(len(outliers))

        if rel_out > 0.1:
            return 'More than 10%% of the samples outside epsilon band for variable "%s"' % name

    return None


def validate_signal(t, y, t_ref, y_ref, t_start, t_stop, num=1000, dx=21, dy=0.1):
    """ Validate a signal y(t) against a reference signal y_ref(t_ref) by creating a band
    around y_ref and finding the values in y outside the band

    Parameters:

        t        time of the signal
        y        values of the signal
        t_ref    time of the reference signal
        y_ref    values of the reference signal
        t_start  start time of the band
        t_stop   stop time of the band
        num      number of samples for the band
        dx       horizontal tolerance in samples
        dy       relative vertical tolerance

    Returns:

        t_band  time values of the band
        y_min   lower limit of the band
        y_max   upper limit of the band
        i_out   indices of the outliers in y
    """

    import numpy as np
    from scipy.ndimage.filters import maximum_filter1d, minimum_filter1d

    # re-sample the reference signal into a uniform grid
    t_band = np.linspace(start=t_start, stop=t_stop, num=num)

    # sort out the duplicate samples before the interpolation
    m = np.concatenate(([True], np.diff(t_ref) > 0))

    y_band = np.interp(x=t_band, xp=t_ref[m], fp=y_ref[m])

    y_band_min = np.min(y_band)
    y_band_max = np.max(y_band)

    # calculate the width of the band
    if y_band_min == y_band_max:
        w = 0.5 if y_band_min == 0 else np.abs(y_band_min) * dy
    else:
        w = (y_band_max - y_band_min) * dy

    # calculate the lower and upper limits
    y_min = minimum_filter1d(input=y_band, size=dx) - w
    y_max = maximum_filter1d(input=y_band, size=dx) + w

    # find outliers
    y_min_i = np.interp(x=t, xp=t_band, fp=y_min)
    y_max_i = np.interp(x=t, xp=t_band, fp=y_max)
    i_out = np.logical_or(y < y_min_i, y > y_max_i)

    # do not count outliers outside the t_ref
    i_out = np.logical_and(i_out, t >= t_start)
    i_out = np.logical_and(i_out, t <= t_stop)

    return t_band, y_min, y_max, i_out



""" This example shows a full factorial variation of two model parameters
using multiple cores.

The goal of this experiment is to determine the losses of a Rectifier for
different AC voltages and DC currents. Therefore the Rectifier.fmu is simulated
for every parameter combination. The losses after 0.02 s (when a steady state has
been reached) are averaged and returned as the result and plotted as a contour plot.
"""

import dask
from dask import bag
from dask.diagnostics import ProgressBar
import numpy as np
import fmpy
from fmpy.fmi2 import FMU2Slave
from fmpy import read_model_description, platform
from fmpy.util import download_test_file
import shutil


fmu_filename = 'Rectifier.fmu'

# define the parameter space for the variation
v_ac = [100, 150, 200, 250, 300, 350, 400, 450, 500]  # AC voltage [V]
i_dc = [20, 30, 40, 50, 60, 70, 80, 90, 100]          # DC current [A]

sync = False  # synchronized execution (for debugging)

stop_time = 0.1

dll_handle = None

n_chunks = 10  # number of chunks to divide the workload

# create the parameters
V_AC, I_DC = np.meshgrid(v_ac, i_dc, indexing='ij')


def simulate_fmu(*args):
    """ Worker function that simulates the FMU

    Parameters:
        args = [indices, fmu_args, start_vrs, result_vrs]

        indices     a list of indices into the parameter arrays
        fmu_args    FMU constructor arguments


    Returns:
        a list of tuples (i, [u_dc, losses]) that contain the index 'i' and the averaged results of the
        'uDC' and 'Losses' variables
    """

    indices, fmu_args, start_vrs, result_vrs = args

    zipped = []

    # global fmu_args, start_vrs, result_vrs, dll_handle

    fmu = FMU2Slave(**fmu_args)

    fmu.instantiate()

    # iterate over the all indices in this batch
    for i in indices:

        # get the start values for the current index
        start_values = [V_AC[i], I_DC[i]]

        fmu.reset()

        fmu.setupExperiment()

        # set the start values
        fmu.setReal(vr=start_vrs, value=start_values)

        fmu.enterInitializationMode()
        fmu.exitInitializationMode()

        time = 0.0
        step_size = 1e-4

        results = []

        # simulation loop
        while time < stop_time:
            fmu.doStep(currentCommunicationPoint=time, communicationStepSize=step_size)
            time += step_size

            if time > 0.02:
                result = fmu.getReal(result_vrs)
                results.append(result)

        u_dc, losses = zip(*results)

        # store the index and the averaged signals
        zipped.append((i, [np.average(u_dc), np.average(losses)]))

    fmu.terminate()

    # call the FMI API directly to avoid unloading the share library
    fmu.fmi2FreeInstance(fmu.component)

    if sync:
        # remember the shared library handle so we can unload it later
        global dll_handle
        dll_handle = fmu.dll._handle
    else:
        # unload the shared library directly
        fmpy.freeLibrary(fmu.dll._handle)

    return zipped


def run_experiment(show_plot=True):

    if platform not in ['win32', 'win64']:
        raise Exception("Rectifier.fmu is only available for Windows")

    print("Parameter variation on %s:" % fmu_filename)
    print("  VAC", v_ac)
    print("  IDC", i_dc)

    if sync:
        dask.config.set(scheduler='synchronous')  # synchronized scheduler

    # download the FMU
    download_test_file('2.0', 'CoSimulation', 'Dymola', '2017', 'Rectifier', fmu_filename)

    # read the model description
    model_description = read_model_description(fmu_filename)

    # collect the value references for the variables to read / write
    vrs = {}
    for variable in model_description.modelVariables:
        vrs[variable.name] = variable.valueReference

    # extract the FMU
    unzipdir = fmpy.extract(fmu_filename)

    fmu_args = {'guid': model_description.guid,
                'modelIdentifier': model_description.coSimulation.modelIdentifier,
                'unzipDirectory': unzipdir}

    # get the value references for the start and output values
    start_vrs = [vrs['VAC'], vrs['IDC']]
    result_vrs = [vrs['uDC'], vrs['Losses']]

    indices = list(np.ndindex(I_DC.shape))

    print("Running %d simulations (%d chunks)..." % (V_AC.size, n_chunks))
    with ProgressBar():
        # calculate the losses for every chunk
        b = bag.from_sequence(indices, npartitions=n_chunks)
        results = b.map_partitions(simulate_fmu, fmu_args, start_vrs, result_vrs).compute()

    LOSSES = np.zeros_like(V_AC)

    # put the results together
    for i, res in results:
        LOSSES[i] = res[1]

    # unload the shared library
    if sync:
        while True:
            try:
                fmpy.freeLibrary(dll_handle)
            except:
                break

    # clean up
    shutil.rmtree(unzipdir, ignore_errors=True)

    if show_plot:
        print("Plotting results...")

        import matplotlib.pyplot as plt

        figure = plt.figure()
        figure.patch.set_facecolor('white')
        ax = figure.add_subplot(1, 1, 1)

        CS = plt.contourf(V_AC, I_DC, LOSSES, 10)
        plt.colorbar(CS, aspect=30)

        CS = ax.contour(V_AC, I_DC, LOSSES, 10, colors='k', linewidths=0.8)
        ax.clabel(CS=CS, fmt='%.0f', fontsize=9, inline=1)

        ax.set_title('Losses / W')
        ax.set_xlabel('AC Voltage / V')
        ax.set_ylabel('DC Current / A')

        plt.show()
    else:
        print("Plotting disabled")

    print("Done.")

    return LOSSES


if __name__ == '__main__':

    run_experiment()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'D:\a\1\s\fmpy\gui\forms\TableDialog.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_TableDialog(object):
    def setupUi(self, TableDialog):
        TableDialog.setObjectName("TableDialog")
        TableDialog.resize(902, 561)
        palette = QtGui.QPalette()
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Window, brush)
        TableDialog.setPalette(palette)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(TableDialog)
        self.verticalLayout_2.setSpacing(15)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.splitter = QtWidgets.QSplitter(TableDialog)
        self.splitter.setStyleSheet("QSplitter::handle:horizontal {\n"
"    border-left: 1px solid #ccc;\n"
"}")
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.tableView = QtWidgets.QTableView(self.splitter)
        self.tableView.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.tableView.setObjectName("tableView")
        self.frame = QtWidgets.QFrame(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame.sizePolicy().hasHeightForWidth())
        self.frame.setSizePolicy(sizePolicy)
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setObjectName("frame")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout.setContentsMargins(0, -1, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.plotSettingsWidget = QtWidgets.QWidget(self.frame)
        self.plotSettingsWidget.setObjectName("plotSettingsWidget")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.plotSettingsWidget)
        self.gridLayout_2.setContentsMargins(20, -1, -1, -1)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.plotRowsRadioButton = QtWidgets.QRadioButton(self.plotSettingsWidget)
        self.plotRowsRadioButton.setObjectName("plotRowsRadioButton")
        self.gridLayout_2.addWidget(self.plotRowsRadioButton, 1, 0, 1, 1)
        self.plotColumnsRadioButton = QtWidgets.QRadioButton(self.plotSettingsWidget)
        self.plotColumnsRadioButton.setChecked(True)
        self.plotColumnsRadioButton.setObjectName("plotColumnsRadioButton")
        self.gridLayout_2.addWidget(self.plotColumnsRadioButton, 0, 0, 1, 1)
        self.firstColumnAsXAxisCheckBox = QtWidgets.QCheckBox(self.plotSettingsWidget)
        self.firstColumnAsXAxisCheckBox.setChecked(True)
        self.firstColumnAsXAxisCheckBox.setObjectName("firstColumnAsXAxisCheckBox")
        self.gridLayout_2.addWidget(self.firstColumnAsXAxisCheckBox, 0, 1, 1, 1)
        self.firstRowAsXAxisCheckBox = QtWidgets.QCheckBox(self.plotSettingsWidget)
        self.firstRowAsXAxisCheckBox.setEnabled(False)
        self.firstRowAsXAxisCheckBox.setChecked(True)
        self.firstRowAsXAxisCheckBox.setObjectName("firstRowAsXAxisCheckBox")
        self.gridLayout_2.addWidget(self.firstRowAsXAxisCheckBox, 1, 1, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_2.addItem(spacerItem, 0, 2, 1, 1)
        self.verticalLayout.addWidget(self.plotSettingsWidget)
        self.graphicsView = GraphicsLayoutWidget(self.frame)
        self.graphicsView.setObjectName("graphicsView")
        self.verticalLayout.addWidget(self.graphicsView)
        self.verticalLayout_2.addWidget(self.splitter)
        self.buttonBox = QtWidgets.QDialogButtonBox(TableDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_2.addWidget(self.buttonBox)

        self.retranslateUi(TableDialog)
        self.buttonBox.accepted.connect(TableDialog.accept)
        self.buttonBox.rejected.connect(TableDialog.reject)
        self.plotColumnsRadioButton.toggled['bool'].connect(self.firstColumnAsXAxisCheckBox.setEnabled)
        self.plotRowsRadioButton.toggled['bool'].connect(self.firstRowAsXAxisCheckBox.setEnabled)
        QtCore.QMetaObject.connectSlotsByName(TableDialog)

    def retranslateUi(self, TableDialog):
        _translate = QtCore.QCoreApplication.translate
        self.plotRowsRadioButton.setText(_translate("TableDialog", "Plot Rows"))
        self.plotColumnsRadioButton.setText(_translate("TableDialog", "Plot Columns"))
        self.firstColumnAsXAxisCheckBox.setText(_translate("TableDialog", "first column as x-axis"))
        self.firstRowAsXAxisCheckBox.setText(_translate("TableDialog", "first row as x-axis"))
from pyqtgraph import GraphicsLayoutWidget
from . import icons_rc



from tempfile import mkdtemp
from typing import List, Tuple
from attr import attrs, attrib, Factory
from ..model_description import Unit, BaseUnit, DisplayUnit, SimpleType


@attrs(eq=False, auto_attribs=True)
class Variable(object):

    type: str = None
    variability: str = None
    causality: str = None
    name: str = None
    start: str = None
    description: str = None
    mapping: Tuple[str, str] = None
    declaredType: str = None
    unit: str = None
    displayUnit: str = None


@attrs(eq=False, auto_attribs=True)
class Component(object):

    filename: str
    name: str


@attrs(eq=False, auto_attribs=True)
class Connection(object):

    startElement: str
    startConnector: str
    endElement: str
    endConnector: str


@attrs(eq=False)
class Configuration(object):

    description = attrib(type=str, default=None, repr=False)
    variableNamingConvention = attrib(type=str, default='flat', repr=False)

    unitDefinitions = attrib(type=List[Unit], default=Factory(list), repr=False)
    typeDefinitions = attrib(type=List[SimpleType], default=Factory(list), repr=False)

    variables = attrib(type=List[Variable], default=Factory(list), repr=False)
    components = attrib(type=List[Component], default=Factory(list), repr=False)
    connections = attrib(type=List[Connection], default=Factory(list), repr=False)


def create_fmu_container(configuration, output_filename):
    """ Create an FMU from nested FMUs (experimental)

        see tests/test_fmu_container.py for an example
    """

    def xml_encode(s):
        """ Escape non-ASCII characters """

        if s is None:
            return ""

        s = s.replace('&', '&amp;')
        s = s.replace('<', '&lt;')
        s = s.replace('>', '&gt;')
        s = s.replace('"', '&quot;')

        for c in s:
            if ord(c) > 127:
                s = s.replace(c, '&#x' + format(ord(c), 'x') + ';')

        return s

    import os
    import shutil
    import fmpy
    from fmpy import read_model_description, extract
    import msgpack
    from datetime import datetime
    import pytz

    base_filename, _ = os.path.splitext(output_filename)
    model_name = os.path.basename(base_filename)

    unzipdir = mkdtemp()

    basedir = os.path.dirname(__file__)

    for directory in ['binaries', 'documentation', 'sources']:
        shutil.copytree(os.path.join(basedir, directory), os.path.join(unzipdir, directory))

    os.mkdir(os.path.join(unzipdir, 'resources'))

    data = {
        'components': [],
        'variables': [],
        'connections': []
    }

    component_map = {}

    for i, component in enumerate(configuration.components):
        model_description = read_model_description(component.filename)
        model_identifier = model_description.coSimulation.modelIdentifier
        extract(component.filename, os.path.join(unzipdir, 'resources', model_identifier))
        variables = dict((v.name, v) for v in model_description.modelVariables)
        component_map[component.name] = (i, variables)
        data['components'].append({
            'name': component.name,
            'guid': model_description.guid,
            'modelIdentifier': model_identifier,
        })

    variables_map = {}

    for i, v in enumerate(configuration.variables):
        variables_map[v.name] = (i, v)

    unit_defintions = ''

    def to_xml(o):

        xml = f'<{type(o).__name__}'

        for a in dir(o):
            if not a.startswith('_'):
                v = getattr(o, a)
                if v:
                    xml += f' {a}="{v}"'

        xml += '/>'

        return xml

    if configuration.unitDefinitions:

        unit_defintions += '\n  <UnitDefinitions>'

        for unit in configuration.unitDefinitions:
            unit_defintions += f'\n    <Unit name="{unit.name}">'
            if unit.baseUnit:
                unit_defintions += '\n      ' + to_xml(unit.baseUnit)
            for displayUnit in unit.displayUnits:
                unit_defintions += '\n      ' + to_xml(displayUnit)
            unit_defintions += '\n    </Unit>'

        unit_defintions += '\n  </UnitDefinitions>'

    type_definitions = ''

    if configuration.typeDefinitions:

        type_definitions += '\n  <TypeDefinitions>'

        for simpleType in configuration.typeDefinitions:
            type_definitions += f'\n    <SimpleType name="{simpleType.name}">'
            type_definitions += f'\n      <{simpleType.type}'
            if simpleType.quantity:
                type_definitions += f' quantity="{simpleType.quantity}'
            if simpleType.unit:
                type_definitions += f' unit="{simpleType.unit}"'
            if simpleType.displayUnit:
                type_definitions += f' displayUnit="{simpleType.displayUnit}"'
            type_definitions += '/>'
            type_definitions += '\n    </SimpleType>'

        type_definitions += '\n  </TypeDefinitions>'

    mv = ''  # model variables
    mo = ''  # model outputs

    for i, v in enumerate(configuration.variables):

        component_indices = []
        value_references = []

        # config.mp
        for component_name, variable_name in v.mapping:
            component_index, component_variables = component_map[component_name]
            value_reference = component_variables[variable_name].valueReference
            component_indices.append(component_index)
            value_references.append(value_reference)

        data['variables'].append({
            'components': component_indices,
            'valueReferences': value_references
        })

        # modelDescription.xml
        start = f' start="{ v.start }"' if v.start else ''
        mv += f'\n    <ScalarVariable name="{ xml_encode(v.name) }" valueReference="{ i }" variability="{ v.variability }" causality="{ v.causality }" description="{ xml_encode(v.description) }">'
        mv += f'\n      <{v.type}{ start }'
        if v.declaredType:
            mv += f' declaredType="{v.declaredType}"'
        if v.unit:
            mv += f' unit="{v.unit}"'
        if v.displayUnit:
            mv += f' displayUnit="{v.displayUnit}"'
        mv += '/>'
        mv += f'\n    </ScalarVariable>'

        if v.causality == 'output':
            mo += f'\n      <Unknown index="{ i + 1 }"/>'

    xml = f'''<?xml version="1.0" encoding="UTF-8"?>
<fmiModelDescription
  fmiVersion="2.0"
  modelName="{ model_name }"
  guid=""
  description="{ xml_encode(configuration.description) }"
  generationTool="FMPy {fmpy.__version__} FMU Container"
  generationDateAndTime="{ datetime.now(pytz.utc).isoformat() }"
  variableNamingConvention="{ configuration.variableNamingConvention }">
  
  <CoSimulation modelIdentifier="FMUContainer">
    <SourceFiles>
      <File name="FMUContainer.c"/>
      <File name="mpack.c"/>
    </SourceFiles>
  </CoSimulation>

  { unit_defintions }

  { type_definitions }

  <ModelVariables>{ mv }
  </ModelVariables>

  <ModelStructure>
    <Outputs>{mo}
    </Outputs>
    <InitialUnknowns>{ mo }
    </InitialUnknowns>
  </ModelStructure>

</fmiModelDescription>
'''

    with open(os.path.join(unzipdir, 'modelDescription.xml'), 'w') as f:
        f.write(xml)

    for c in configuration.connections:
        data['connections'].append({
            'type': component_map[c.startElement][1][c.startConnector].type,
            'startComponent': component_map[c.startElement][0],
            'endComponent': component_map[c.endElement][0],
            'startValueReference': component_map[c.startElement][1][c.startConnector].valueReference,
            'endValueReference': component_map[c.endElement][1][c.endConnector].valueReference,
        })

    with open(os.path.join(unzipdir, 'resources', 'config.mp'), 'wb') as f:
        packed = msgpack.packb(data)
        f.write(packed)

    shutil.make_archive(base_filename, 'zip', unzipdir)

    if os.path.isfile(output_filename):
        os.remove(output_filename)

    os.rename(base_filename + '.zip', output_filename)

    shutil.rmtree(unzipdir, ignore_errors=True)



def generate_result_tables(repo_dir, data_dir):
    """ Generate the cross-check result tables """

    import os
    import csv
    from fmpy.cross_check import get_vendor_ids

    combinations = []  # all permutations of FMI version, type and platform

    for fmi_version in ['1.0', '2.0']:
        for fmi_type in ['cs', 'me']:
            for platform in ['c-code', 'darwin64', 'linux32', 'linux64', 'win32', 'win64']:
                combinations.append((fmi_version, fmi_type, platform))

    tools_csv = os.path.join(data_dir, 'tools.csv')

    vendors = get_vendor_ids(tools_csv)

    tools = {}  # tool_id -> tool_name

    for tool_infos in vendors.values():
        for tool_id, tool_name in tool_infos:
            tools[tool_id] = tool_name

    def split_path(path):

        segments = []

        while True:
            path, segment = os.path.split(path)
            if not segment:
                break
            segments.insert(0, segment)

        return segments

    def collect_results():

        results = []

        vendor_repo = os.path.join(repo_dir, 'results')

        for root, dirs, files in os.walk(vendor_repo):

            if 'passed' not in files:
                continue  # ignore

            segments = split_path(root)

            fmi_version, fmi_type, platform, _, _, exporting_tool_name, exporting_tool_version, model_name = segments[-8:]

            not_compliant_file = os.path.join(repo_dir, 'fmus', fmi_version, fmi_type, platform, exporting_tool_name, exporting_tool_version, model_name, 'notCompliantWithLatestRules')

            if os.path.isfile(not_compliant_file):
                continue  # ignore

            results.append(segments[-8:])

        return results

    def build_matrix(results, fmi_version, fmi_type, platform):
        """ Build the result matrix for an FMI version, type and platform """

        importing_tools = set()
        exporting_tools = set()

        filtered = []

        # get the tools
        for fmi_version_, fmi_type_, platform_, importing_tool_name, importing_tool_version, exporting_tool_name, exporting_tool_version, model_name in results:

            if fmi_version_ != fmi_version or fmi_type_ != fmi_type or platform_ != platform:
                continue

            importing_tools.add(importing_tool_name)
            exporting_tools.add(exporting_tool_name)

            filtered.append((importing_tool_name, importing_tool_version, exporting_tool_name, exporting_tool_version, model_name))

        # build matrix
        importing_tools = sorted(importing_tools, key=lambda s: s.lower())
        exporting_tools = sorted(exporting_tools, key=lambda s: s.lower())

        matrix = []

        for importing_tool in importing_tools:
            row = []
            for exporting_tool in exporting_tools:
                count = 0
                for r in filtered:
                    if r[0] == importing_tool and r[2] == exporting_tool:
                        count += 1
                row.append(count)
            matrix.append(row)

        return importing_tools, exporting_tools, matrix

    results = collect_results()

    # filter tool IDs
    results = [r for r in results if r[3] in tools and r[5] in tools]

    matrices = {}

    for combination in combinations:
        matrices[combination] = build_matrix(results, *combination)

    for fmi_version, fmi_type, platform in combinations:

        importing_tools, exporting_tools, matrix = matrices[(fmi_version, fmi_type, platform)]

        importing_tools = [tools[tool_id] for tool_id in importing_tools]
        exporting_tools = [tools[tool_id] for tool_id in exporting_tools]

        csv_filename = 'fmi1' if fmi_version == '1.0' else 'fmi2'
        csv_filename += '-'
        csv_filename += fmi_type
        csv_filename += '-'
        csv_filename += platform + '.csv'

        with open(os.path.join(data_dir, 'cross-check', csv_filename), 'w') as f:
            f.write(','.join([''] + exporting_tools) + '\n')
            for importing_tool, row in zip(importing_tools, matrix):
                f.write(','.join([importing_tool] + list(map(str, row))) + '\n')

    participants = set()

    # generate the tools file with cross-check results
    export_passed = {}
    import_passed = {}

    for fmi_version, fmi_type, platform, importing_tool_name, importing_tool_version, exporting_tool_name, exporting_tool_version, model_name in results:

        key = (exporting_tool_name, fmi_version, fmi_type)

        participants.add(exporting_tool_name)
        participants.add(importing_tool_name)

        if key not in export_passed:
            export_passed[key] = {}

        if importing_tool_name not in export_passed[key]:
            export_passed[key][importing_tool_name] = {model_name}
        else:
            export_passed[key][importing_tool_name].add(model_name)

        key = (importing_tool_name, fmi_version, fmi_type)

        if key not in import_passed:
            import_passed[key] = {}

        if exporting_tool_name not in import_passed[key]:
            import_passed[key][exporting_tool_name] = {model_name}
        else:
            import_passed[key][exporting_tool_name].add(model_name)

    # aggregate the results
    participants = sorted(participants, key=lambda s: s.lower())

    rows = [participants]

    def check_passed(key, d):
        cnt = 0
        for vendor, models in d.get(key, {}).items():
            if len(models) >= 3:
                cnt += 1
        return cnt >= 3

    for results in [export_passed, import_passed]:

        for i, c in enumerate([('cs', '1.0'), ('cs', '2.0'), ('me', '1.0'), ('me', '2.0')]):

            fmi_type, fmi_version = c
            row = []

            for tool_id in participants:
                key = (tool_id, fmi_version, fmi_type)
                row.append('passed' if check_passed(key, results) else '')

            rows.append(row)

    with open(os.path.join(data_dir, 'cross-check', 'result.csv'), 'w', newline='') as f:
        writer = csv.writer(f)
        writer.writerows(rows)


if __name__ is '__main__':

    import argparse

    parser = argparse.ArgumentParser(description="Generate the cross-check result tables")

    parser.add_argument('xc_repo_dir', help="Cloned cross-check repository")
    parser.add_argument('data_dir', help="_data directory in the fmi-standard.org repository")

    args = parser.parse_args()

    generate_result_tables(repo_dir=args.xc_repo_dir, data_dir=args.data_dir)



from .libraries import sundials_cvode
from .sundials_types import *
from .sundials_nvector import *

# /* -----------------------------------------------------------------
#  * Programmer(s): Scott D. Cohen, Alan C. Hindmarsh, Radu Serban,
#  *                and Dan Shumaker @ LLNL
#  * -----------------------------------------------------------------
#  * SUNDIALS Copyright Start
#  * Copyright (c) 2002-2019, Lawrence Livermore National Security
#  * and Southern Methodist University.
#  * All rights reserved.
#  *
#  * See the top-level LICENSE and NOTICE files for details.
#  *
#  * SPDX-License-Identifier: BSD-3-Clause
#  * SUNDIALS Copyright End
#  * -----------------------------------------------------------------
#  * This is the header file for the main CVODE integrator.
#  * -----------------------------------------------------------------*/
#
# #ifndef _CVODE_H
# #define _CVODE_H
#
# #include <stdio.h>
# #include <sundials/sundials_nvector.h>
# #include <sundials/sundials_nonlinearsolver.h>
# #include <cvode/cvode_ls.h>
#
# #ifdef __cplusplus  /* wrapper to enable C++ usage */
# extern "C" {
# #endif
#
# /* -----------------
#  * CVODE Constants
#  * ----------------- */
#
# /* lmm */
# #define CV_ADAMS          1
CV_ADAMS = 1
# #define CV_BDF            2
CV_BDF = 2
#
# /* itask */
# #define CV_NORMAL         1
CV_NORMAL = 1
# #define CV_ONE_STEP       2
#
#
# /* return values */
#
# #define CV_SUCCESS               0
CV_SUCCESS = 0
# #define CV_TSTOP_RETURN          1
# #define CV_ROOT_RETURN           2
CV_ROOT_RETURN = 2
#
# #define CV_WARNING              99
#
# #define CV_TOO_MUCH_WORK        -1
# #define CV_TOO_MUCH_ACC         -2
# #define CV_ERR_FAILURE          -3
# #define CV_CONV_FAILURE         -4
#
# #define CV_LINIT_FAIL           -5
# #define CV_LSETUP_FAIL          -6
# #define CV_LSOLVE_FAIL          -7
# #define CV_RHSFUNC_FAIL         -8
# #define CV_FIRST_RHSFUNC_ERR    -9
# #define CV_REPTD_RHSFUNC_ERR    -10
# #define CV_UNREC_RHSFUNC_ERR    -11
# #define CV_RTFUNC_FAIL          -12
# #define CV_NLS_INIT_FAIL        -13
# #define CV_NLS_SETUP_FAIL       -14
# #define CV_CONSTR_FAIL          -15
# #define CV_NLS_FAIL             -16
#
# #define CV_MEM_FAIL             -20
# #define CV_MEM_NULL             -21
# #define CV_ILL_INPUT            -22
# #define CV_NO_MALLOC            -23
# #define CV_BAD_K                -24
# #define CV_BAD_T                -25
# #define CV_BAD_DKY              -26
# #define CV_TOO_CLOSE            -27
# #define CV_VECTOROP_ERR         -28
#
# #define CV_UNRECOGNIZED_ERR     -99
#
#
# /* ------------------------------
#  * User-Supplied Function Types
#  * ------------------------------ */
#
# typedef int (*CVRhsFn)(realtype t, N_Vector y,
#                        N_Vector ydot, void *user_data);
CVRhsFn = CFUNCTYPE(c_int, realtype, N_Vector, N_Vector, c_void_p)
#
# typedef int (*CVRootFn)(realtype t, N_Vector y, realtype *gout,
#                         void *user_data);
CVRootFn = CFUNCTYPE(c_int, realtype, N_Vector, POINTER(realtype), c_void_p)
#
# typedef int (*CVEwtFn)(N_Vector y, N_Vector ewt, void *user_data);
#
# typedef void (*CVErrHandlerFn)(int error_code,
#                                const char *module, const char *function,
#                                char *msg, void *user_data);
CVErrHandlerFn = CFUNCTYPE(None, c_int, c_char_p, c_char_p, c_char_p, c_void_p)
#
# /* -------------------
#  * Exported Functions
#  * ------------------- */
#
# /* Initialization functions */
# SUNDIALS_EXPORT void *CVodeCreate(int lmm);
CVodeCreate = getattr(sundials_cvode, 'CVodeCreate')
CVodeCreate.argtypes = [c_int]
CVodeCreate.restype = c_void_p
#
# SUNDIALS_EXPORT int CVodeInit(void *cvode_mem, CVRhsFn f, realtype t0,
#                               N_Vector y0);
CVodeInit = getattr(sundials_cvode, 'CVodeInit')
CVodeInit.argtypes = [c_void_p, CVRhsFn, realtype, N_Vector]
CVodeInit.restype = c_int
# SUNDIALS_EXPORT int CVodeReInit(void *cvode_mem, realtype t0, N_Vector y0);
CVodeReInit = getattr(sundials_cvode, 'CVodeReInit')
CVodeReInit.argtypes = [c_void_p, realtype, N_Vector]
CVodeReInit.restype = c_int
#
# /* Tolerance input functions */
# SUNDIALS_EXPORT int CVodeSStolerances(void *cvode_mem, realtype reltol,
#                                       realtype abstol);
# SUNDIALS_EXPORT int CVodeSVtolerances(void *cvode_mem, realtype reltol,
#                                       N_Vector abstol);
CVodeSVtolerances = getattr(sundials_cvode, 'CVodeSVtolerances')
CVodeSVtolerances.argtypes = [c_void_p, realtype, N_Vector]
CVodeSVtolerances.restype = c_int
# SUNDIALS_EXPORT int CVodeWFtolerances(void *cvode_mem, CVEwtFn efun);
#
# /* Optional input functions */
# SUNDIALS_EXPORT int CVodeSetErrHandlerFn(void *cvode_mem, CVErrHandlerFn ehfun,
#                                          void *eh_data);
CVodeSetErrHandlerFn = getattr(sundials_cvode, 'CVodeSetErrHandlerFn')
CVodeSetErrHandlerFn.argtypes = [c_void_p, CVErrHandlerFn, c_void_p]
CVodeSetErrHandlerFn.restype = c_int
# SUNDIALS_EXPORT int CVodeSetErrFile(void *cvode_mem, FILE *errfp);
# SUNDIALS_EXPORT int CVodeSetUserData(void *cvode_mem, void *user_data);
# SUNDIALS_EXPORT int CVodeSetMaxOrd(void *cvode_mem, int maxord);
# SUNDIALS_EXPORT int CVodeSetMaxNumSteps(void *cvode_mem, long int mxsteps);
CVodeSetMaxNumSteps = getattr(sundials_cvode, 'CVodeSetMaxNumSteps')
CVodeSetMaxNumSteps.argtypes = [c_void_p, c_long]
CVodeSetMaxNumSteps.restype = c_int
# SUNDIALS_EXPORT int CVodeSetMaxHnilWarns(void *cvode_mem, int mxhnil);
# SUNDIALS_EXPORT int CVodeSetStabLimDet(void *cvode_mem, booleantype stldet);
# SUNDIALS_EXPORT int CVodeSetInitStep(void *cvode_mem, realtype hin);
# SUNDIALS_EXPORT int CVodeSetMinStep(void *cvode_mem, realtype hmin);
# SUNDIALS_EXPORT int CVodeSetMaxStep(void *cvode_mem, realtype hmax);
CVodeSetMaxStep = getattr(sundials_cvode, 'CVodeSetMaxStep')
CVodeSetMaxStep.argtypes = [c_void_p, realtype]
CVodeSetMaxStep.restype = c_int
# SUNDIALS_EXPORT int CVodeSetStopTime(void *cvode_mem, realtype tstop);
# SUNDIALS_EXPORT int CVodeSetMaxErrTestFails(void *cvode_mem, int maxnef);
# SUNDIALS_EXPORT int CVodeSetMaxNonlinIters(void *cvode_mem, int maxcor);
# SUNDIALS_EXPORT int CVodeSetMaxConvFails(void *cvode_mem, int maxncf);
# SUNDIALS_EXPORT int CVodeSetNonlinConvCoef(void *cvode_mem, realtype nlscoef);
# SUNDIALS_EXPORT int CVodeSetConstraints(void *cvode_mem, N_Vector constraints);
#
# SUNDIALS_EXPORT int CVodeSetNonlinearSolver(void *cvode_mem,
#                                             SUNNonlinearSolver NLS);
#
# /* Rootfinding initialization function */
# SUNDIALS_EXPORT int CVodeRootInit(void *cvode_mem, int nrtfn, CVRootFn g);
CVodeRootInit = getattr(sundials_cvode, 'CVodeRootInit')
CVodeRootInit.argtypes = [c_void_p, c_int, CVRootFn]
CVodeRootInit.restype = c_int
#
# /* Rootfinding optional input functions */
# SUNDIALS_EXPORT int CVodeSetRootDirection(void *cvode_mem, int *rootdir);
# SUNDIALS_EXPORT int CVodeSetNoInactiveRootWarn(void *cvode_mem);
CVodeSetNoInactiveRootWarn = getattr(sundials_cvode, 'CVodeSetNoInactiveRootWarn')
CVodeSetNoInactiveRootWarn.argtypes = [c_void_p]
CVodeSetNoInactiveRootWarn.restype = c_int
#
# /* Solver function */
# SUNDIALS_EXPORT int CVode(void *cvode_mem, realtype tout, N_Vector yout,
#                           realtype *tret, int itask);
CVode = getattr(sundials_cvode, 'CVode')
CVode.argtypes = [c_void_p, realtype, N_Vector, POINTER(realtype), c_int]
CVode.restype = c_int
#
# /* Dense output function */
# SUNDIALS_EXPORT int CVodeGetDky(void *cvode_mem, realtype t, int k,
#                                 N_Vector dky);
#
# /* Optional output functions */
# SUNDIALS_EXPORT int CVodeGetWorkSpace(void *cvode_mem, long int *lenrw,
#                                       long int *leniw);
# SUNDIALS_EXPORT int CVodeGetNumSteps(void *cvode_mem, long int *nsteps);
# SUNDIALS_EXPORT int CVodeGetNumRhsEvals(void *cvode_mem, long int *nfevals);
# SUNDIALS_EXPORT int CVodeGetNumLinSolvSetups(void *cvode_mem,
#                                              long int *nlinsetups);
# SUNDIALS_EXPORT int CVodeGetNumErrTestFails(void *cvode_mem,
#                                             long int *netfails);
# SUNDIALS_EXPORT int CVodeGetLastOrder(void *cvode_mem, int *qlast);
# SUNDIALS_EXPORT int CVodeGetCurrentOrder(void *cvode_mem, int *qcur);
# SUNDIALS_EXPORT int CVodeGetCurrentGamma(void *cvode_mem, realtype *gamma);
# SUNDIALS_EXPORT int CVodeGetNumStabLimOrderReds(void *cvode_mem,
#                                                 long int *nslred);
# SUNDIALS_EXPORT int CVodeGetActualInitStep(void *cvode_mem, realtype *hinused);
# SUNDIALS_EXPORT int CVodeGetLastStep(void *cvode_mem, realtype *hlast);
# SUNDIALS_EXPORT int CVodeGetCurrentStep(void *cvode_mem, realtype *hcur);
# SUNDIALS_EXPORT int CVodeGetCurrentState(void *cvode_mem, N_Vector *y);
# SUNDIALS_EXPORT int CVodeGetCurrentTime(void *cvode_mem, realtype *tcur);
# SUNDIALS_EXPORT int CVodeGetTolScaleFactor(void *cvode_mem, realtype *tolsfac);
# SUNDIALS_EXPORT int CVodeGetErrWeights(void *cvode_mem, N_Vector eweight);
# SUNDIALS_EXPORT int CVodeGetEstLocalErrors(void *cvode_mem, N_Vector ele);
# SUNDIALS_EXPORT int CVodeGetNumGEvals(void *cvode_mem, long int *ngevals);
# SUNDIALS_EXPORT int CVodeGetRootInfo(void *cvode_mem, int *rootsfound);
CVodeGetRootInfo = getattr(sundials_cvode, 'CVodeGetRootInfo')
CVodeGetRootInfo.argtypes = [c_void_p, POINTER(c_int)]
CVodeGetRootInfo.restype = c_int
# SUNDIALS_EXPORT int CVodeGetIntegratorStats(void *cvode_mem, long int *nsteps,
#                                             long int *nfevals,
#                                             long int *nlinsetups,
#                                             long int *netfails,
#                                             int *qlast, int *qcur,
#                                             realtype *hinused, realtype *hlast,
#                                             realtype *hcur, realtype *tcur);
# SUNDIALS_EXPORT int CVodeGetNumNonlinSolvIters(void *cvode_mem,
#                                                long int *nniters);
# SUNDIALS_EXPORT int CVodeGetNumNonlinSolvConvFails(void *cvode_mem,
#                                                    long int *nncfails);
# SUNDIALS_EXPORT int CVodeGetNonlinSolvStats(void *cvode_mem, long int *nniters,
#                                             long int *nncfails);
# SUNDIALS_EXPORT char *CVodeGetReturnFlagName(long int flag);
#
# /* Free function */
# SUNDIALS_EXPORT void CVodeFree(void **cvode_mem);
CVodeFree = getattr(sundials_cvode, 'CVodeFree')
CVodeFree.argtypes = [POINTER(c_void_p)]
CVodeFree.restype = None
#
#
# #ifdef __cplusplus
# }
# #endif
#
# #endif



""" Interface to the SUNDIALS libraries """

import numpy as np
from .cvode import *
from .cvode_ls import *
from .nvector_serial import *
from .sundials_linearsolver import *
from .sundials_matrix import *
from .sundials_nvector import *
from .sundials_types import *
from .sundials_version import *
from .sunlinsol_dense import *
from .sunmatrix_dense import *


def _assert_cv_success(status):
    if status != CV_SUCCESS:
        raise Exception("Call to CVode failed.")


def _assert_version():
    major = c_int()
    minor = c_int()
    patch = c_int()
    len = 8
    label = create_string_buffer(len)
    _assert_cv_success(SUNDIALSGetVersionNumber(byref(major), byref(minor), byref(patch), label, len))
    if major.value != 5:
        raise Exception("Wrong SUNDIALS major version number. Expected 5 but was %d." % major.value)


_assert_version()


class CVodeSolver(object):
    """ Interface to the CVode solver """

    def __init__(self,
                 nx, nz, get_x, set_x, get_dx, get_z, set_time, input,
                 startTime,
                 maxStep=float('inf'),
                 relativeTolerance=1e-5,
                 maxNumSteps=500):
        """
        Parameters:
            nx                  number of continuous states
            nz                  number of event indicators
            get_x               callback function to get the continuous states
            set_x               callback function to set the continuous states
            get_dx              callback function to get the derivatives
            get_z               callback function to get the event indicators
            set_time            callback function to set the time
            startTime           start time for the integration
            maxStep             maximum absolute value of step size allowed
            relativeTolerance   relative tolerance
            maxNumSteps         maximum number of internal steps to be taken by the solver in its attempt to reach tout
        """

        self.get_x = get_x
        self.set_x = set_x
        self.get_dx = get_dx
        self.get_z = get_z
        self.set_time = set_time
        self.input = input
        self.error_info = None

        self.discrete = nx == 0

        if self.discrete:
            # insert a dummy state
            self.nx = 1
        else:
            self.nx = nx

        self.nz = nz

        self.x      = N_VNew_Serial(self.nx)
        self.abstol = N_VNew_Serial(self.nx)

        self.px      = NV_DATA_S(self.x)
        self.pabstol = NV_DATA_S(self.abstol)

        # initialize
        if self.discrete:
            x = np.ctypeslib.as_array(self.px, (self.nx,))
            x[:] = 1.0
        else:
            self.get_x(self.px, self.nx)

        abstol = np.ctypeslib.as_array(self.pabstol, (self.nx,))
        abstol[:] = relativeTolerance

        self.cvode_mem = CVodeCreate(CV_BDF)

        # add function pointers as members to save them from GC
        self.f_ = CVRhsFn(self.f)
        self.g_ = CVRootFn(self.g)
        self.ehfun_ = CVErrHandlerFn(self.ehfun)

        _assert_cv_success(CVodeInit(self.cvode_mem, self.f_, startTime, self.x))

        _assert_cv_success(CVodeSVtolerances(self.cvode_mem, relativeTolerance, self.abstol))

        _assert_cv_success(CVodeRootInit(self.cvode_mem, self.nz, self.g_))

        self.A = SUNDenseMatrix(self.nx, self.nx)

        self.LS = SUNLinSol_Dense(self.x, self.A)

        _assert_cv_success(CVodeSetLinearSolver(self.cvode_mem, self.LS, self.A))

        _assert_cv_success(CVodeSetMaxStep(self.cvode_mem, maxStep))

        _assert_cv_success(CVodeSetMaxNumSteps(self.cvode_mem, maxNumSteps))

        _assert_cv_success(CVodeSetNoInactiveRootWarn(self.cvode_mem))

        _assert_cv_success(CVodeSetErrHandlerFn(self.cvode_mem, self.ehfun_, None))

    def ehfun(self, error_code, module, function, msg,  user_data):
        """ Error handler function """
        self.error_info = (error_code, module.decode("utf-8"), function.decode("utf-8"), msg.decode("utf-8"))

    def f(self, t, y, ydot, user_data):
        """ Right-hand-side function """

        self.set_time(t)

        if self.discrete:
            dx = np.ctypeslib.as_array(NV_DATA_S(ydot), (self.nx,))
            dx[:] = 0.0
        else:
            self.set_x(NV_DATA_S(y), self.nx)
            self.get_dx(NV_DATA_S(ydot), self.nx)
        return 0

    def g(self, t, y, gout, user_data):
        """ Root function """

        self.set_time(t)
        self.input.apply(t)

        if not self.discrete:
            self.set_x(NV_DATA_S(y), self.nx)

        self.get_z(gout, self.nz)

        return 0

    def step(self, t, tNext):

        if not self.discrete:
            # get the states
            self.get_x(self.px, self.nx)

        tret = realtype(0.0)

        # perform one step
        flag = CVode(self.cvode_mem, tNext, self.x, byref(tret), CV_NORMAL)

        if not self.discrete:
            # set the states
            self.set_x(self.px, self.nx)

        roots_found = np.zeros(shape=(self.nz,), dtype=np.int32)

        if flag == CV_ROOT_RETURN:
            p_roots_found = np.ctypeslib.as_ctypes(roots_found)
            _assert_cv_success(CVodeGetRootInfo(self.cvode_mem, p_roots_found))
        elif flag < 0:
            raise RuntimeError("CVode error (code %s) in module %s, function %s: %s" % self.error_info)

        return flag > 0, roots_found, tret.value

    def reset(self, time):

        if not self.discrete:
            self.get_x(self.px, self.nx)

        # reset the solver
        flag = CVodeReInit(self.cvode_mem, time, self.x)

    def __del__(self):
        # clean up
        CVodeFree(byref(c_void_p(self.cvode_mem)))



""" This example demonstrates how to save CPU time by reusing the extracted FMU,
 loaded model description, and FMU instance when simulating the same FMU multiple times """

from fmpy import *
from fmpy.util import download_file
import shutil


def run_efficient_loop():

    # download the FMU
    url = 'https://github.com/modelica/fmi-cross-check/raw/master/fmus/2.0/cs/win64/Test-FMUs/0.0.2/VanDerPol/VanDerPol.fmu'
    sha = 'a870f5f7f712e8152bfd60a1c2fd1c0bc10d4ca8124bd3031e321e8dd1e71bb0'
    filename = download_file(url, sha)

    # extract the FMU to a temporary directory
    unzipdir = extract(filename)

    # read the model description
    model_description = read_model_description(unzipdir)

    # instantiate the FMU
    fmu_instance = instantiate_fmu(unzipdir, model_description, 'ModelExchange')

    # perform the iteration
    for i in range(100):

        # reset the FMU instance instead of creating a new one
        fmu_instance.reset()

        # calculate the parameters for this run
        start_values = {'mu': i * 0.01}

        # pass the unzipdir, model description and FMU instance to simulate_fmu()
        result = simulate_fmu(unzipdir,
                              start_values=start_values,
                              model_description=model_description,
                              fmu_instance=fmu_instance)

    # free the FMU instance and unload the shared library
    fmu_instance.freeInstance()

    # delete the temporary directory
    shutil.rmtree(unzipdir, ignore_errors=True)


if __name__ == '__main__':
    run_efficient_loop()



import os
import shutil
import numpy as np

from fmpy import read_model_description, extract
from fmpy.fmi1 import FMU1Slave
from fmpy.fmi2 import FMU2Slave
from fmpy.ssp.ssd import System, read_ssd, get_connections, find_connectors, find_components


def get_value(component, name):
    """ Get a variable from a component """

    variable = component.variables[name]
    vr = [variable.valueReference]

    if variable.type == 'Real':
        return component.fmu.getReal(vr)[0]
    elif variable.type in ['Integer', 'Enumeration']:
        return component.fmu.getInteger(vr)[0]
    elif variable.type == 'Boolean':
        value = component.fmu.getBoolean(vr)[0]
        return value != 0
    else:
        raise Exception("Unsupported type: %s" % variable.type)


def set_value(component, name, value):
    """ Set a variable to a component """

    variable = component.variables[name]
    vr = [variable.valueReference]

    if variable.type == 'Real':
        component.fmu.setReal(vr, [float(value)])
    elif variable.type in ['Integer', 'Enumeration']:
        component.fmu.setInteger(vr, [int(value)])
    elif variable.type == 'Boolean':
        # TODO: convert literals
        component.fmu.setBoolean(vr, [value != 0.0])
    else:
        raise Exception("Unsupported type: %s" % variable.type)


def add_path(element, path=''):

    if isinstance(element, System):
        for child in element.elements:
            add_path(child, path + child.name + '.')

    for connector in element.connectors:
        connector.path = path + connector.name


def set_parameters(component, parameter_set):
    """ Apply the parameters (start values) to a component """

    path = component.name

    parent = component.parent

    while parent.parent is not None:
        path = parent.name + '.' + path
        parent = parent.parent

    for parameter in parameter_set.parameters:
        if parameter.name.startswith(path):
            variable_name = parameter.name[len(path) + 1:]
            set_value(component, variable_name, parameter.value)


def instantiate_fmu(component, ssp_unzipdir, start_time, stop_time=None, parameter_set=None):
    """ Instantiate an FMU """

    fmu_filename = os.path.join(ssp_unzipdir, component.source)

    component.unzipdir = extract(fmu_filename)

    # read the model description
    model_description = read_model_description(fmu_filename, validate=False)

    if model_description.coSimulation is None:
        raise Exception("%s does not support co-simulation." % component.source)

    # collect the value references
    component.variables = {}
    for variable in model_description.modelVariables:
        # component.vrs[variable.name] = variable.valueReference
        component.variables[variable.name] = variable

    fmu_kwargs = {'guid': model_description.guid,
                  'unzipDirectory': component.unzipdir,
                  'modelIdentifier': model_description.coSimulation.modelIdentifier,
                  'instanceName': component.name}

    if model_description.fmiVersion == '1.0':
        component.fmu = FMU1Slave(**fmu_kwargs)
        component.fmu.instantiate()
        if parameter_set is not None:
            set_parameters(component, parameter_set)
        component.fmu.initialize(stopTime=stop_time)
    else:
        component.fmu = FMU2Slave(**fmu_kwargs)
        component.fmu.instantiate()
        component.fmu.setupExperiment(startTime=start_time)
        if parameter_set is not None:
            set_parameters(component, parameter_set)
        component.fmu.enterInitializationMode()
        component.fmu.exitInitializationMode()


def free_fmu(component):
    """ Free an FMU and remove its unzip dir """

    component.fmu.terminate()
    component.fmu.freeInstance()
    try:
        shutil.rmtree(component.unzipdir)
    except Exception as e:
        print("Failed to remove unzip directory. " + str(e))


def do_step(component, time, step_size):
    """ Perform one simulation step """

    # set inputs
    for connector in component.connectors:
        if connector.kind == 'input':
            set_value(component, connector.name, connector.value)

    # do step
    component.fmu.doStep(currentCommunicationPoint=time, communicationStepSize=step_size)

    # get outputs
    for connector in component.connectors:
        if connector.kind == 'output':
            connector.value = get_value(component, connector.name)


def simulate_ssp(ssp_filename, start_time=0.0, stop_time=None, step_size=None, parameter_set=None, input={}):
    """ Simulate a system of FMUs """

    if stop_time is None:
        stop_time = 1.0

    if step_size is None:
        step_size = stop_time * 1e-2

    ssd = read_ssd(ssp_filename)

    add_path(ssd.system)

    components = find_components(ssd.system)
    connectors = find_connectors(ssd.system)
    connections = get_connections(ssd.system)

    # resolve connections
    connections_reversed = {}

    for a, b in connections:
        connections_reversed[b] = a

    new_connections = []

    # trace connections back to the actual start connector
    for a, b in connections:

        while isinstance(a.parent, System) and a.parent.parent is not None:
            a = connections_reversed[a]

        new_connections.append((a, b))

    connections = new_connections

    # extract the SSP
    ssp_unzipdir = extract(ssp_filename)

    # initialize the connectors
    for connector in connectors:
        connector.value = 0.0

    # instantiate the FMUs
    for component in components:
        instantiate_fmu(component, ssp_unzipdir, start_time, stop_time, parameter_set)

    time = start_time

    rows = []  # list to record the results

    # simulation loop
    while time < stop_time:

        # apply input
        for connector in ssd.system.connectors:
            if connector.kind == 'input' and connector.name in input:
                connector.value = input[connector.name](time)

        # perform one step
        for component in components:
            do_step(component, time, step_size)

        # apply connections
        for start_connector, end_connector in connections:
            end_connector.value = start_connector.value

        # get the results
        row = [time]

        for connector in connectors:
            row.append(connector.value)

        # append the results
        rows.append(tuple(row))

        # advance the time
        time += step_size

    # free the FMUs
    for component in components:
        free_fmu(component)

    # clean up
    shutil.rmtree(ssp_unzipdir)

    dtype = [('time', np.float64)]

    for connector, value in zip(connectors, rows[0][1:]):
        if type(value) == bool:
            dtype.append((connector.path, np.bool_))
        elif type(value) == int:
            dtype.append((connector.path, np.int32))
        else:
            dtype.append((connector.path, np.float64))

    # convert the results to a structured NumPy array
    return np.array(rows, dtype=np.dtype(dtype))
